from .path_alg import solve_path, pathalgo_general
import numpy as np
import numpy.linalg as LA
from .solve_R1 import problem_R1, Classo_R1

"""
Problem    :   min h_rho(Ab - y) + lambda ||b||1 with C.b= 0 <=>   min ||Ab - y - r*o||^2 + lambda ||b,o||1 with C.b= 0, o in R^m
                                                                                        r = lambda / 2rho
Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda.
The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type',
which is defined bellow in order to contain all the important parameters of the problem.
One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo).
We solve the problem without normalizing anything.
"""


def Classo_R2(pb, lam, compute=True):

    pb_type = pb.type  # can be 'Path-Alg', 'P-PDS' , 'PF-PDS' or 'DR'

    (m, d, k), (A, C, y) = pb.dim, pb.matrix
    lamb, rho = lam * pb.lambdamax, pb.rho

    if lam == 0.0:
        pb_type = "DR"
        compute = "True"
        # here we simply refer to Classo_R1 that is called line 42.

    # Path-Alg
    # here we compute the path algo until our lambda, and just take the last beta

    if pb_type == "Path-Alg":
        if pb.intercept:
            AA, CC = A[:, 1:], C[:, 1:]
        else:
            AA, CC = A[:, :], C[:, :]
        out = solve_path((AA, CC, y), lam, False, rho, "R2", intercept=pb.intercept)
        if pb.intercept:
            beta0, beta = out[0][-1], out[1][-1]
            beta = np.array([beta0] + list(beta))
        else:
            beta = out[0][-1]

        return beta

    # DR :
    regpath = pb.regpath
    r = lamb / (2 * rho)
    if pb_type == "DR":
        if compute:
            pb.init_R1(r=r)
            x = Classo_R1(pb.prob_R1, lamb / pb.prob_R1.lambdamax)
            beta = x[:-m]
            if pb.intercept:
                betaO = pb.prob_R1.ybar - np.vdot(pb.prob_R1.Abar, x)
                beta = np.array([betaO] + list(beta))
            return beta
        else:
            pb.add_r(r=r)
            if len(pb.init) == 3:
                pb.prob_R1.init = pb.init
            x, warm_start = Classo_R1(pb.prob_R1, lamb / pb.prob_R1.lambdamax)
            beta = x[:-m]
            if pb.intercept:
                betaO = pb.prob_R1.ybar - np.vdot(pb.prob_R1.Abar, x)
                beta = np.array([betaO] + list(beta))
            return (beta, warm_start)

    tol = pb.tol * LA.norm(y) / LA.norm(A, "fro")  # tolerance rescaled

    # cvx
    # call to the cvx function of minimization
    if pb_type == "cvx":
        import cvxpy as cp

        x = cp.Variable(d)
        objective, constraints = (
            cp.Minimize(cp.sum(cp.huber(A * x - y, rho)) + lamb * cp.norm(x, 1)),
            [C * x == 0],
        )
        prob = cp.Problem(objective, constraints)
        result = prob.solve(warm_start=regpath, eps_abs=tol)
        if regpath:
            return (x.value, True)
        return x.value

    if compute:
        pb.compute_param()
    tau, Proj, AtA, Aty = pb.tauN, proj_c(C, d), pb.AtA, pb.Aty
    gamma = pb.gam / (2 * (pb.AtAnorm + r ** 2))
    t = lamb * gamma
    w, tm, zerom, zerod = (
        t * pb.weights,
        t * np.ones(m),
        np.zeros(m),
        np.zeros(d),
    )
    o, xbar, x, v = pb.init
    # vectors usefull to compute the prox of f(b)= sum(wi |bi|)

    # FORWARD BACKWARD
    if pb_type == "P-PDS":

        for i in range(pb.N):
            grad = AtA.dot(x) - Aty
            v = v + tau * C.dot(xbar)
            S = x - 2 * gamma * grad - 2 * gamma * r * (A.T).dot(o) - (C.T).dot(v)
            o = prox(
                o * (1 - 2 * gamma * r ** 2) + 2 * gamma * r * (y - A.dot(x)),
                tm,
                zerom,
            )
            p = prox(S, w, zerod)
            nw_x = Proj.dot(p)
            eps = nw_x - x
            xbar = p + eps

            if i % 10 == 2 and LA.norm(eps) < tol:  # 0.6
                if regpath:
                    return (x, (o, xbar, x, v))
                else:
                    return x
            x = nw_x
            if LA.norm(x) > 1e10:
                raise ValueError("The algorithm of P-PDS diverges")

        raise ValueError(
            "The algorithm of P-PDS did not converge after %i iterations " % pb.N
        )

    # NO PROJ

    if pb_type == "PF-PDS":  # y1 --> S ; p1 --> p . ; p2 --> y2
        for i in range(pb.N):
            grad = AtA.dot(x) - Aty

            S1 = x - 2 * gamma * grad - 2 * gamma * r * (A.T).dot(o) - (C.T).dot(v)
            S2 = o * (1 - 2 * gamma * r ** 2) + 2 * gamma * r * (y - A.dot(x))

            p1 = prox(S1, w, zerod)
            p2 = prox(S2, tm, zerom)

            v = v + tau * C.dot(p1)
            v2 = v + tau * C.dot(x)

            eps1 = (
                p1 + 2 * gamma * (Aty - AtA.dot(p1) - r * A.T.dot(o)) - C.T.dot(v2) - S1
            )
            eps2 = p2 + 2 * r * gamma * (y - r * p2 - A.dot(x)) - S2

            x = x + eps1
            o = o + eps2

            if i > 0 and LA.norm(eps1) + LA.norm(eps2) < tol:
                if regpath:
                    return (x, (o, xbar, x, v))
                else:
                    return x

            if LA.norm(x) + LA.norm(o) + LA.norm(v) > 1e6:
                raise ValueError("The algorithm of PF-PDS diverges")

        raise ValueError(
            "The algorithm of PF-PDS did not converge after %i iterations " % pb.N
        )


"""
This function compute the the solution for a given path of lam :
by calling the function 'algo' for each lambda with warm start,
or wuth the method ODE, by computing the whole path thanks to the ODE that rules Beta and the subgradient s,
and then to evaluate it in the given finite path.
"""


def pathlasso_R2(pb, path, n_active=False):
    n = pb.dim[0]
    BETA, tol = [], pb.tol
    if pb.type == "Path-Alg":
        (A, C, y) = pb.matrix
        if pb.intercept:
            AA, CC = A[:, 1:], C[:, 1:]
        else:
            AA, CC = A[:, :], C[:, :]
        return pathalgo_general((AA, CC, y), path, "R2", n_active, pb.rho, pb.intercept)

    # Now we are in the case where we have to do warm starts.
    save_init = pb.init
    pb.regpath = True
    pb.compute_param()
    pb.init_R1()
    if type(n_active) == int and n_active > 0:
        n_act = n_active
    else:
        n_act = n
    for lam in path:
        X = Classo_R2(pb, lam, compute=False)
        BETA.append(X[0])
        pb.init = X[1]
        if (
            sum([(abs(X[0][i]) > 1e-1) for i in range(len(X[0]))]) >= n_act
            or type(X[1]) == str
        ):
            pb.init = save_init
            BETA.extend([BETA[-1]] * (len(path) - len(BETA)))
            pb.regpath = False
            return BETA

    pb.init = save_init
    pb.regpath = False
    return BETA


"""
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
"""


class problem_R2:
    def __init__(self, data, algo, rho, intercept=False):
        self.N = 500000

        (AA, CC, y) = data
        A = AA[:, :]
        C = CC[:, :]
        self.weights = np.ones(A.shape[1])
        self.intercept = intercept
        if intercept:
            # add a column of 1 in A, and change weight.
            A = np.concatenate([np.ones((len(A), 1)), A], axis=1)
            C = np.concatenate([np.zeros((len(C), 1)), C], axis=1)
            self.weights = np.concatenate([[0.0], self.weights])
            yy = y - np.mean(y)

        self.dim = (
            A.shape[0],
            A.shape[1],
            C.shape[0],
        )
        self.matrix = (A, C, y)
        (m, d, k) = self.dim
        self.init = np.zeros(m), np.zeros(d), np.zeros(d), np.zeros(k)
        self.tol = 1e-4
        self.regpath = False
        self.name = algo + " Huber"
        self.type = algo  # type of algo used
        self.rho = rho
        self.gam = 1.0
        self.tau = 0.5  # equation for the convergence of Noproj and LS algorithms : gam + tau < 1
        if not intercept:
            yy = y
        self.lambdamax = 2 * LA.norm(AA.T.dot(h_prime(yy, rho)), np.infty)

    """
    this is a method of the class pb that is used to computed the expensive multiplications only once. (espacially usefull for warm start. )
    """

    def compute_param(self):
        (A, C, y) = self.matrix
        m, d, k = self.dim
        self.c = (
            d / LA.norm(A, 2)
        ) ** 2  # parameter for Concomitant problem : the matrix is scaled as c*A^2

        self.AtA = (A.T).dot(A)
        self.Aty = (A.T).dot(y)
        self.Cnorm = LA.norm(C, 2) ** 2
        self.tauN = self.tau / self.Cnorm
        self.AtAnorm = LA.norm(self.AtA, 2)

    def init_R1(self, r=0.0):
        (AA, CC, y) = self.matrix
        A, C = AA[:, :], CC[:, :]
        (m, d, k) = self.dim
        if self.intercept:
            A, C = A[:, 1:], C[:, 1:]

        Ahuber = np.append(A, r * np.eye(m), 1)
        Chuber = np.append(C, np.zeros((k, m)), 1)
        yhuber = y
        if self.intercept:
            Abar = np.mean(Ahuber, axis=0)
            ybar = np.mean(y)
            Ahuber = Ahuber - Abar
            yhuber = yhuber - ybar
        matrices_huber = (Ahuber, Chuber, yhuber)
        prob = problem_R1(matrices_huber, self.type)
        prob.regpath = self.regpath
        prob.compute_param()
        if self.intercept:
            prob.Abar = Abar
            prob.ybar = ybar
            self.AAt = (A - np.mean(A, axis=0)).dot((A - np.mean(A, axis=0)).T)
        else:
            self.AAt = A.dot(A.T)
        self.prob_R1 = prob

    def add_r(self, r):
        prob = self.prob_R1
        A_r1 = prob.matrix[0]
        m = A_r1.shape[0]
        d = A_r1.shape[1] - m
        prob.AtA[d:, :d] = prob.matrix[0][:, :d] * r
        prob.AtA[:d, d:] = prob.AtA[d:, :d].T
        prob.Aty = np.append(prob.Aty[:d], prob.matrix[2] * r)
        prob.lambdamax = 2 * LA.norm(prob.Aty, np.infty)
        extension = np.eye(m)
        if self.intercept:
            # self.init_R1(r=r)
            extension = extension - np.mean(extension, axis=0)
        extension = r * extension
        A_r1[:, d:] = extension
        right_bottom = extension.dot(extension)
        prob.AtA[d:, d:] = right_bottom
        prob.AAt = self.AAt + right_bottom
        prob.AtAnorm = LA.norm(prob.AtA, 2)


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b, w, zeros):
    return np.minimum(b + w, zeros) + np.maximum(b - w, zeros)


# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M, d):
    if LA.matrix_rank(M) == 0:
        return np.eye(d)
    return np.eye(d) - LA.multi_dot([M.T, np.linalg.inv(M.dot(M.T)), M])


# Compute the derivative of the huber function, particulary useful for the computing of lambdamax
def h_prime(y, rho):
    m = len(y)
    lrho = rho * np.ones(m)
    return np.maximum(lrho, -y) + np.minimum(y - lrho, 0)



"""
The package is organized as follow :
    There is a main class called :obj:`classo_problem`, that contains a lot of information about the problem,
    and once the problem is solved, it will also contains the solution.

    Here is the global structure of the problem instance:

        A :obj:`classo_problem` instance contains a :obj:`Data` instance, a :obj:`Formulation` instance, a :obj:`Model_selection` instance and a :obj:`Solution` instance.


        A :obj:`Model_selection` instance contains the instances : :obj:`PATHparameters`, :obj:`CVparameters`, :obj:`StabSelparameters`, :obj:`LAMfixedparameters`.


        A :obj:`Solution` instance, once is computed, contains the instances : :obj:`solution_PATH`, :obj:`solution_CV`, :obj:`solution_StabSel`, :obj:`solution_LAMfixed`.

"""


from time import time
import numpy as np
import matplotlib.pyplot as plt

from .misc_functions import (
    rescale,
    theoretical_lam,
    min_LS,
    affichage,
    check_size,
    tree_to_matrix,
)
from .compact_func import Classo, pathlasso
from .cross_validation import CV
from .stability_selection import stability, selected_param
import matplotlib.patches as mpatches


class classo_problem:
    """Class that contains all the information about the problem.
    It also has a representation method so one can print it.

    Args:
        X (ndarray): Matrix representing the data of the problem.
        y (ndarray): Vector representing the output of the problem.
        C (str or ndarray, optional ): Matrix of constraints to the problem. If it is 'zero-sum' then the corresponding attribute will be all-one matrix.
            Default value : 'zero-sum'
        rescale (bool, optional): if True, then the function :func:`rescale` will be applied to data when solving the problem.
            Default value : False
        label (list,optional) : list of the labels of each variable. If None, then label are just indices.
            Default value : None

    Attributes:
        data (Data) :  object containing the data (matrices) of the problem. Namely : X, y, C and the labels.
        formulation (Formulation) : object containing the info about the formulation of the minimization problem we solve.
        model_selection (Model_selection) : object containing the parameters we need to do variable selection.
        solution (Solution) : object giving caracteristics of the solution of the model_selection that is asked.
            Before using the method :func:`solve()` , its componant are empty/null.
        numerical_method (str) : name of the numerical method that is used, it can be :
            'Path-Alg' (path algorithm) , 'P-PDS' (Projected primal-dual splitting method) , 'PF-PDS' (Projection-free primal-dual splitting method) or 'DR' (Douglas-Rachford-type splitting method).
            Default value : 'not specified', which means that the function :func:`choose_numerical_method` will choose it accordingly to the formulation.

    """

    def __init__(
        self, X, y, C=None, Tree=None, label=None, rescale=False
    ):  # zero sum constraint by default, but it can be any matrix
        self.data = Data(X, y, C, Tree=Tree, rescale=rescale, label=label)
        self.formulation = Formulation()
        self.model_selection = Model_selection()
        self.solution = Solution()
        self.numerical_method = "not specified"

    # This method is the way to solve the model selections contained in the object Model_selection, with the formulation of 'formulation' and the data.
    def solve(self):
        """Method that solves every model required in the attributes of the problem instance
        and update the attribute :attr:`solution` with the characteristics of the solution."""
        data = self.data
        self.solution = Solution()
        matrices = (data.X, data.C, data.y)
        n, d = len(data.X), len(data.X[0])
        if self.formulation.classification:
            self.formulation.concomitant = False

        if type(self.formulation.e) == str:
            if self.formulation.e == "n/2":
                self.formulation.e = (
                    n / 2
                )  # useful to be able to write e='n/2' as it is in the default parameters
            elif self.formulation.e == "n":
                self.formulation.e = n  # same
            else:
                if self.formulation.huber:
                    self.formulation.e = n
                else:
                    self.formulation.e = n / 2

        if data.rescale:
            matrices, data.scaling = rescale(matrices)  # SCALING contains  :
            # (list of initial norms of A-colomns,
            #         initial norm of centered y,
            #          mean of initial y )

        if self.formulation.w is not None:
            if min(self.formulation.w) < 1e-8:
                raise ValueError(
                    "w has to be positive weights, here it has a value smaller than 1e-8"
                )

        if len(data.label) > d:
            sup = len(data.label) - d
            data.label = data.label[sup:]
            print(
                "too many labels, there for the labels {} have been deleted".format(
                    data.label[:sup]
                )
            )

        elif len(data.label) < d:
            missing = d - len(data.label)
            print(
                " too few labels, therefore {} labels have been inserted in the end".format(
                    missing
                )
            )
            data.label = np.array(
                list(data.label) + ["missing " + str(i) for i in range(missing)]
            )

        if self.formulation.intercept:
            data.label = np.array(["intercept"] + list(data.label))

        label = data.label

        # Compute the path thanks to the class solution_path which contains directely the computation in the initialisation
        if self.model_selection.PATH:
            self.solution.PATH = solution_PATH(
                matrices,
                self.model_selection.PATHparameters,
                self.formulation,
                self.numerical_method,
                label,
            )

        # Compute the cross validation thanks to the class solution_CV which contains directely the computation in the initialisation
        if self.model_selection.CV:
            self.solution.CV = solution_CV(
                matrices,
                self.model_selection.CVparameters,
                self.formulation,
                self.numerical_method,
                label,
            )

        # Compute the Stability Selection thanks to the class solution_SS which contains directely the computation in the initialisation
        if self.model_selection.StabSel:
            param = self.model_selection.StabSelparameters
            param.theoretical_lam = theoretical_lam(int(n * param.percent_nS), d)
            if not param.rescaled_lam:
                param.theoretical_lam = param.theoretical_lam * int(
                    n * param.percent_nS
                )

            self.solution.StabSel = solution_StabSel(
                matrices, param, self.formulation, self.numerical_method, label
            )

        # Compute the c-lasso problem at a fixed lam thanks to the class solution_LAMfixed which contains directely the computation in the initialisation
        if self.model_selection.LAMfixed:
            param = self.model_selection.LAMfixedparameters
            param.theoretical_lam = theoretical_lam(n, d)
            if not param.rescaled_lam:
                param.theoretical_lam = param.theoretical_lam * n
            self.solution.LAMfixed = solution_LAMfixed(
                matrices, param, self.formulation, self.numerical_method, label
            )

    def __repr__(self):
        print_parameters = ""
        if self.model_selection.LAMfixed:
            print_parameters += (
                "\n \nLAMBDA FIXED PARAMETERS: "
                + self.model_selection.LAMfixedparameters.__repr__()
            )

        if self.model_selection.PATH:
            print_parameters += (
                "\n \nPATH PARAMETERS: "
                + self.model_selection.PATHparameters.__repr__()
            )

        if self.model_selection.CV:
            print_parameters += (
                "\n \nCROSS VALIDATION PARAMETERS: "
                + self.model_selection.CVparameters.__repr__()
            )

        if self.model_selection.StabSel:
            print_parameters += (
                "\n \nSTABILITY SELECTION PARAMETERS: "
                + self.model_selection.StabSelparameters.__repr__()
            )

        return (
            " \n \nFORMULATION: "
            + self.formulation.__repr__()
            + "\n \n"
            + "MODEL SELECTION COMPUTED:  "
            + self.model_selection.__repr__()
            + print_parameters
            + "\n"
        )


class Data:
    """Class that contains the data of the problem
    ie where matrices and labels are stored.

    Args:
        X (ndarray): Matrix representing the data of the problem.
        y (ndarray): Vector representing the output of the problem.
        C (str or array, optional ): Matrix of constraints to the problem. If it is 'zero-sum' then the corresponding attribute will be all-one matrix.
        rescale (bool, optional): if True, then the function :func:`rescale` will be applied to data when solving the problem.
            Default value : False
        label (list, optional) : list of the labels of each variable. If None, then labels are juste the indices.
            Default value : None
        Tree (skbio.TreeNode, optional) : taxonomic tree, if not None, then the matrices X and C and the labels will be changed.

    Attributes:
        X (ndarray): Matrix representing the data of the problem.
        y (ndarray): Vector representing the output of the problem.
        C (str or array, optional ): Matrix of constraints to the problem. If it is 'zero-sum' then the corresponding attribute will be all-one matrix.
        rescale (bool, optional): if True, then the function :func:`rescale` will be applied to data when solving the problem.
        label (list) : list of the labels of each variable. If None, then labels are juste the indices.
        tree (skbio.TreeNode or None) : taxonomic tree.

    """

    def __init__(self, X, y, C, Tree=None, rescale=False, label=None):
        self.rescale = rescale  # booleen to know if we rescale the matrices
        X1, y1, C1 = check_size(X, y, C)

        if Tree is None:
            if label is None:
                self.label = np.array([str(i) for i in range(len(X[0]))])
            else:
                self.label = np.array(label)
            self.X, self.y, self.C, self.tree = X1, y1, C1, None

        else:
            A, label2, subtree = tree_to_matrix(Tree, label, with_repr=True)
            self.tree = subtree
            self.X, self.y, self.C, self.label = (
                X1.dot(A),
                y1,
                C1.dot(A),
                np.array(label2),
            )


class Formulation:
    """Class that contains the information about the formulation of the problem
    namely, the type of formulation (R1, R2, R3, R4, C1, C2)
    and its parameters like rho, the weigths and the presence of an intercept.
    The type of formulation is encoded with boolean huber concomitant and classification
    with the rule:

        False False False = R1

        True False False = R2

        False True False = R3

        True True False = R4

        False False True = C1

        True False True = C2

    It also has a representation method so one can print it.

    Attributes:
        huber (bool) : True if the formulation of the problem should be robust.
            Default value : False

        concomitant (bool) : True if the formulation of the problem should be with an M-estimation of sigma.
            Default value : True

        classification (bool) : True if the formulation of the problem should be classification (if yes, then it will not be concomitant).
            Default value : False

        rho (float) : Value of rho for R2 and R4 formulations.
            Default value : 1.345

        rho_classification (float) : value of rho for huberized hinge loss function for classification ie C2 (it has to be strictly smaller then 1).
            Default value : 0.

        e (float or string)  : value of e in concomitant formulation.
            If 'n/2' then it becomes n/2 during the method :func:`solve()`, same for 'n'.
            Default value : 'n' if huber formulation ; 'n/2' else

        w (numpy ndarray) : array of size d with the weights of the L1 penalization. This has to be positive.
            Default value : None (which makes it the 1,...,1 vector)

        intercept (bool)  : set to true if we should use an intercept.
            Default value : False

    """

    def __init__(self):
        self.huber = False
        self.concomitant = True
        self.classification = False
        self.rho = 1.345
        self.rho_classification = 0.0
        self.e = "not specified"
        self.w = None
        self.intercept = False

    def name(self):
        if self.classification:
            if self.huber:
                return "C2"
            else:
                return "C1"
        if self.concomitant:
            if self.huber:
                return "R4"
            else:
                return "R3"
        if self.huber:
            return "R2"
        else:
            return "R1"

    def __repr__(self):
        return self.name()


class Model_selection:
    """Class that contains information about the model selections to perform.
    It contains boolean that states which one will be computed.
    It also contains objects that contain parameters of each computation modes.
    It also has a representation method so one can print it.

    Attributes:
        PATH (bool): True if path should be computed.
            Default value : False

        PATHparameters (PATHparameters): object containing parameters to compute the lasso-path.

        CV (bool):  True if Cross Validation should be computed.
            Default value : False

        CVparameters (CVparameters):  object containing parameters to compute the cross-validation.


        StabSel (boolean):  True if Stability Selection should be computed.
            Default value : True

        StabSelparameters (StabSelparameters):  object containing parameters to compute the stability selection.

        LAMfixed (boolean):  True if solution for a fixed lambda should be computed.
            Default value : False

        LAMfixedparameters (LAMfixedparameters):  object containing parameters to compute the lasso for a fixed lambda.

    """

    def __init__(self, method="not specified"):

        # Model selection variables

        self.PATH = False
        self.PATHparameters = PATHparameters(method=method)

        self.CV = False
        self.CVparameters = CVparameters(method=method)

        self.StabSel = True  # Only model selection that is used by default
        self.StabSelparameters = StabSelparameters(method=method)

        self.LAMfixed = False
        self.LAMfixedparameters = LAMfixedparameters(method=method)

    def __repr__(self):
        string = ""
        if self.LAMfixed:
            string += "\n     Lambda fixed"
        if self.PATH:
            string += "\n     Path"
        if self.CV:
            string += "\n     Cross Validation"
        if self.StabSel:
            string += "\n     Stability selection"

        return string


class PATHparameters:
    """Class that contains the parameters to compute the lasso-path.
    It also has a representation method so one can print it.

    Attributes:
        numerical_method (str) : name of the numerical method that is used, it can be :
            'Path-Alg' (path algorithm) , 'P-PDS' (Projected primal-dual splitting method),
            'PF-PDS' (Projection-free primal-dual splitting method) or 'DR' (Douglas-Rachford-type splitting method).
            Default value : 'not specified', which means that the function :func:`choose_numerical_method` will choose it accordingly to the formulation

        n_active (int): if it is higher than 0, then the algo stops computing the path when n_active variables are active.
        Then the solution does not change from this point.
            Default value : 0

        lambdas (numpy.ndarray) : list of rescaled lambdas for computing lasso-path.
            Default value : None, which means line space between 1 and :attr:`lamin` and :attr:`Nlam` points, with logarithm scale or not depending on :attr:`logscale`.

        Nlam (int) : number of points in the lambda-path if :attr:`lambdas` is still None (default).
            Default value : 80

        lamin (float) : lambda minimum if :attr:`lambdas` is still None (default).
            Default value : 1e-3

        logscale (bool): when :attr:`lambdas` is set to None (default), this parameters tells if it should be set with log scale or not.
            Default value : True

        plot_sigma (bool) : if True then the representation method of the solution will also plot the sigma-path if it is computed (formulation R3 or R4).
            Default value : True

        label (numpy.ndarray of str) : labels on each coefficient.

    """

    def __init__(self, method="not specified"):
        self.formulation = "not specified"
        self.numerical_method = method
        self.n_active = 0
        self.Nlam = 80
        self.lamin = 1e-3
        self.logscale = True
        self.lambdas = None
        self.plot_sigma = True

    def __repr__(self):
        if self.lambdas is not None:
            self.Nlam = len(self.lambdas)
            self.lamin = min(self.lambdas)
            typ = " "
        else:
            if self.logscale:
                typ = " with log-scale"
            else:
                typ = " with linear-scale"

        string = "\n     numerical_method : " + str(self.numerical_method)
        string += "\n     lamin = " + str(self.lamin)
        string += "\n     Nlam = " + str(self.Nlam)
        string += "\n" + typ
        if self.n_active > 0:
            string += "\n     maximum active variables = " + str(self.n_active)

        return string


class CVparameters:
    """Class that contains the parameters to compute the cross-validation.
    It also has a representation method so one can print it.

    Attributes:
        seed (bool or int, optional) : Seed for random values, for an equal seed, the result will be the same. If set to False/None: pseudo-random seed.
            Default value : 0

        numerical_method (str) : name of the numerical method that is used, can be :
            'Path-Alg' (path algorithm) , 'P-PDS' (Projected primal-dual splitting method),
            'PF-PDS' (Projection-free primal-dual splitting method) or 'DR' (Douglas-Rachford-type splitting method).
            Default value : 'not specified', which means that the function :func:`choose_numerical_method` will choose it accordingly to the formulation.

        lambdas (numpy.ndarray) : list of rescaled lambdas for computing lasso-path.
            Default value : None, which means line space between 1 and :attr:`lamin` and :attr:`Nlam` points, with logarithm scale or not depending on :attr:`logscale`.

        Nlam (int) : number of points in the lambda-path if :attr:`lambdas` is still None (default).
            Default value : 80

        lamin (float) : lambda minimum if :attr:`lambdas` is still None (default).
            Default value : 1e-3

        logscale (bool): when :attr:`lambdas` is set to None (default), this parameters tells if it should be set with log scale or not.
            Default value : True

        oneSE (bool) : if set to True, the selected lambda is computed with method 'one-standard-error'.
            Default value : True

        Nsubset (int): number of subset in the cross validation method.
            Default value : 5

    """

    def __init__(self, method="not specified"):
        self.seed = 0
        self.formulation = "not specified"
        self.numerical_method = method
        self.Nsubset = 5  # Number of subsets used
        self.Nlam = 80
        self.lamin = 1e-3
        self.logscale = True
        self.lambdas = None
        self.oneSE = True

    def __repr__(self):
        if self.lambdas is not None:
            self.Nlam = len(self.lambdas)
            self.lamin = min(self.lambdas)
            typ = " "
        else:
            if self.logscale:
                typ = " with log-scale"
            else:
                typ = " with linear-scale"

        string = "\n     numerical_method : " + str(self.numerical_method)
        string += "\n     one-SE method : " + str(self.oneSE)
        string += "\n     Nsubset = " + str(self.Nsubset)
        string += "\n     lamin = " + str(self.lamin)
        string += "\n     Nlam = " + str(self.Nlam)
        string += "\n" + typ

        return string


class StabSelparameters:
    """Class that contains the parameters to compute the stability selection.
    It also has a representation method so one can print it.

    Attributes:

        seed (bool or int, optional) : Seed for random values, for an equal seed, the result will be the same. If set to False/None: pseudo-random seed.
            Default value : 123

        numerical_method (str) : name of the numerical method that is used, can be :
            'Path-Alg' (path algorithm) , 'P-PDS' (Projected primal-dual splitting method) , 'PF-PDS' (Projection-free primal-dual splitting method) or 'DR' (Douglas-Rachford-type splitting method).
            Default value : 'not specified', which means that the function :func:`choose_numerical_method` will choose it accordingly to the formulation.

        lam (float or str) : (only used if :obj:`method` = 'lam') lam for which the lasso should be computed.
            Default value : 'theoretical' which mean it will be equal to :obj:`theoretical_lam` once it is computed.

        rescaled_lam (bool) : (only used if :obj:`method` = 'lam') False if lam = lambda, False if lam = lambda/lambdamax which is between 0 and 1.
            If False and lam = 'theoretical' , then it will take the value n*theoretical_lam.
            Default value : True


        theoretical_lam (float) : (only used if :obj:`method` = 'lam') Theoretical lam.
            Default value : 0.0 (once it is not computed yet, it is computed thanks to the function :func:`theoretical_lam` used in :meth:`classo_problem.solve`).


        method (str) : 'first', 'lam' or 'max' depending on the type of stability selection we do.
            Default value : 'first'

        B (int) : number of subsample considered.
            Default value : 50

        q (int) : number of selected variable per subsample.
            Default value : 10

        percent_nS (float) : size of subsample relatively to the total amount of sample.
            Default value : 0.5

        lamin (float) : lamin when computing the lasso-path for method 'max'.
            Default value : 1e-2

        hd (bool) : if set to True, then the 'max' will stop when it reaches n-k actives variables.
            Default value : False

        threshold (float) : threshold for stability selection.
            Default value : 0.7

        threshold_label (float) : threshold to know when the label should be plot on the graph.
            Default value : 0.4

    """

    def __init__(self, method="not specified"):
        self.seed = 123
        self.formulation = "not specified"
        self.numerical_method = method

        self.method = "first"  # Can be 'first' ; 'max' or 'lam'
        self.B = 50
        self.q = 10
        self.percent_nS = 0.5
        self.Nlam = 50  # for path computation
        self.lamin = 1e-2  # the lambda where one stop for 'max' method
        self.hd = False  # if set to True, then the 'max' will stop when it reaches n-k actives variables
        self.lam = "theoretical"  # can also be a float, for the 'lam' method
        self.rescaled_lam = True
        self.threshold = 0.7
        self.threshold_label = 0.4
        self.theoretical_lam = 0.0

    def __repr__(self):
        string = "\n     numerical_method : " + str(self.numerical_method)
        string += "\n     method : " + str(self.method)
        string += "\n     B = " + str(self.B)
        string += "\n     q = " + str(self.q)
        string += "\n     percent_nS = " + str(self.percent_nS)
        string += "\n     threshold = " + str(self.threshold)

        if self.method == "lam":
            string += "\n     lam = " + str(self.lam)
            if self.theoretical_lam != 0.0:
                string += "\n     theoretical_lam = " + str(
                    round(self.theoretical_lam, 4)
                )
        else:
            string += "\n     lamin = " + str(self.lamin)
            string += "\n     Nlam = " + str(self.Nlam)

        return string


class LAMfixedparameters:
    """Class that contains the parameters to compute the lasso for a fixed lambda.
    It also has a representation method so one can print it.

    Attributes:
        numerical_method (str) : name of the numerical method that is used, can be :
            'Path-Alg' (path algorithm) , 'P-PDS' (Projected primal-dual splitting method) , 'PF-PDS' (Projection-free primal-dual splitting method) or 'DR' (Douglas-Rachford-type splitting method).
            Default value : 'not specified', which means that the function :func:`choose_numerical_method` will choose it accordingly to the formulation

        lam (float or str) : lam for which the lasso should be computed.
            Default value : 'theoretical' which mean it will be equal to :obj:`theoretical_lam` once it is computed

        rescaled_lam (bool) : False if lam = lambda, True if lam = lambda/lambdamax which is between 0 and 1.
            If False and lam = 'theoretical' , then it will takes the  value n*theoretical_lam.
            Default value : True

        theoretical_lam (float) : Theoretical lam.
            Default value : 0.0 (once it is not computed yet, it is computed thanks to the function :func:`theoretical_lam` used in :meth:`classo_problem.solve`).

        threshold (float) : Threshold such that the parameters i selected or the ones such as the absolute value of beta[i] is greater than the threshold.
            If None, then it will be set to the average of the absolute value of beta.
            Default value : None
    """

    def __init__(self, method="not specified"):
        self.lam = "theoretical"
        self.formulation = "not specified"
        self.numerical_method = method
        self.rescaled_lam = True
        self.theoretical_lam = 0.0
        self.threshold = None

    def __repr__(self):
        string = "\n     numerical_method = " + str(self.numerical_method)
        string += "\n     rescaled lam : " + str(self.rescaled_lam)
        string += "\n     threshold = " + str(round(self.threshold, 3))
        if type(self.lam) is str:
            string += "\n     lam : " + self.lam
        else:
            string += "\n     lam = " + str(round(self.lam, 3))

        if self.theoretical_lam != 0.0:
            string += "\n     theoretical_lam = " + str(round(self.theoretical_lam, 4))

        return string


class Solution:
    """Class that contains  characteristics of the solution of the model_selections that are computed
    Before using the method :func:`solve()` , its componant are empty/null.
    It also has a representation method so one can print it.

    Attributes:
        PATH (solution_PATH): Solution components of the model PATH.
        CV (solution_CV):  Solution components of the model CV.
        StabelSel (solution_StabSel): Solution components of the model StabSel.
        LAMfixed (solution_LAMfixed): Solution components of the model LAMfixed.

    """

    def __init__(self):
        self.PATH = "not computed"  # this will be filled with an object of the class 'solution_PATH' when the method solve() will be used.
        self.CV = "not computed"  # will be an object of the class 'solution_PATH'
        self.StabSel = (
            "not computed"  # will be an object of the class 'solution_StabSel'
        )
        self.LAMfixed = "not computed"

    def __repr__(self):
        string = ""
        if not type(self.LAMfixed) is str:
            string += self.LAMfixed.__repr__() + "\n"
        if not type(self.PATH) is str:
            string += self.PATH.__repr__() + "\n"
        if not type(self.CV) is str:
            string += self.CV.__repr__() + "\n"
        if not type(self.StabSel) is str:
            string += self.StabSel.__repr__() + "\n"

        return string


# Here, the main function used is pathlasso ; from the file compact_func
class solution_PATH:
    """Class that contains  characteristics of the lasso-path computed,
    which also contains representation method that plot the graphic of this lasso-path.

    Attributes:
        BETAS (numpy.ndarray) : array of size Npath x d with the solution beta for each lambda on each row.
        SIGMAS (numpy.ndarray) : array of size Npath with the solution sigma for each lambda when the formulation of the problem is R2 or R4.
        LAMBDAS (numpy.ndarray) : array of size Npath with the lambdas (real lambdas, not divided by lambda_max) for which the solution is computed.
        method (str) : name of the numerical method that has been used. It can be 'Path-Alg', 'P-PDS' , 'PF-PDS' or 'DR'.
        save (bool or str) : if it is a str, then it gives the name of the file where the graphics has been/will be saved (after using print(solution) ).
        formulation (Formulation) : object containing the info about the formulation of the minimization problem we solve.
        time (float) : running time of this action.

    """

    def __init__(self, matrices, param, formulation, numerical_method, label):
        t0 = time()

        # Formulation choosing
        if param.formulation == "not specified":
            param.formulation = formulation
        if param.numerical_method == "not specified":
            param.numerical_method = numerical_method
        name_formulation = param.formulation.name()
        rho = param.formulation.rho
        rho_classification = param.formulation.rho_classification
        e = param.formulation.e
        # Algorithmic method choosing
        numerical_method = choose_numerical_method(
            param.numerical_method, "PATH", param.formulation
        )
        param.numerical_method = numerical_method
        # Compute the solution and is the formulation is concomitant, it also compute sigma
        if param.lambdas is None:
            if param.logscale:
                param.lambdas = np.array(
                    [param.lamin ** (i / (param.Nlam - 1)) for i in range(param.Nlam)]
                )
            else:
                np.linspace(1.0, param.lamin, param.Nlam)

        out = pathlasso(
            matrices,
            lambdas=param.lambdas,
            n_active=param.n_active,
            typ=name_formulation,
            meth=numerical_method,
            return_sigm=True,
            rho=rho,
            e=e,
            rho_classification=rho_classification,
            w=param.formulation.w,
            intercept=param.formulation.intercept,
        )
        if formulation.concomitant:
            self.BETAS, self.LAMBDAS, self.SIGMAS = out
        else:
            self.BETAS, self.LAMBDAS = out
            self.SIGMAS = "not computed"

        self.formulation = formulation
        self.plot_sigma = param.plot_sigma
        self.method = numerical_method
        self.save = False
        self.label = label
        self.time = time() - t0

    def __repr__(self):

        string = "\n PATH COMPUTATION : "
        d = len(self.BETAS[0])

        if (
            d > 20
        ):  # this trick is to plot only the biggest value, excluding the intercept
            avg_betas = np.mean(abs(np.array(self.BETAS)), axis=0)
            if self.formulation.intercept:
                avg_betas[0] = 0  # trick to exclude intercept in the graph
                string += "\n There is also an intercept.  "
            top = np.argpartition(avg_betas, -20)[-20:]

        else:
            if self.formulation.intercept:
                top = np.arange(1, d)
                string += "\n There is also an intercept.  "
            else:
                top = np.arange(d)

        affichage(
            self.BETAS[:, top],
            self.LAMBDAS,
            labels=self.label[top],
            naffichage=5,
            title=PATH_beta_path["title"] + self.formulation.name(),
            xlabel=PATH_beta_path["xlabel"],
            ylabel=PATH_beta_path["ylabel"],
        )
        if type(self.save) == str:
            plt.savefig(self.save + "Beta-path")
        plt.show()
        if type(self.SIGMAS) != str and self.plot_sigma:
            plt.plot(self.LAMBDAS, self.SIGMAS), plt.ylabel(
                PATH_sigma_path["ylabel"]
            ), plt.xlabel(PATH_sigma_path["xlabel"])
            plt.title(PATH_sigma_path["title"] + self.formulation.name())
            if type(self.save) == str:
                plt.savefig(self.save + "Sigma-path")
            plt.show()

        string += "\n   Running time :  " + str(round(self.time, 3)) + "s"
        return string


# Here, the main function used is CV ; from the file cross_validation
class solution_CV:
    """Class that contains  characteristics of the cross validation computed,
    which also contains a representation method
    that plot the selected parameters and the solution of the not-sparse problem on the selected variables set.

    Attributes:
        xGraph (numpy.ndarray) : array of size Nlam of the lambdas / lambda_max.
        yGraph (numpy.ndarray) : array of size Nlam of the average validation residual (over the K subsets).
        standard_error (numpy.ndarray) : array of size Nlam of the standard error of the validation residual (over the K subsets).
        index_min (int) : index on xGraph of the selected lambda without 1-standard-error method.
        index_1SE (int) : index on xGraph of the selected lambda with 1-standard-error method.
        lambda_min (float) : selected lambda without 1-standard-error method.
        lambda_oneSE (float) : selected lambda with 1-standard-error method.
        beta (numpy.ndarray) : solution beta of classo at lambda_oneSE/lambda_min depending on :attr:`CVparameters.oneSE`.
        sigma (float) : solution sigma of classo at lambda_oneSE when formulation is 'R2' or 'R4'.
        selected_param (numpy.ndarray) : boolean arrays of size d with True when the variable is selected.
        refit (numpy.ndarray) : solution beta after solving unsparse problem over the set of selected variables.
        save1,save2 (bool or string) : if a string is given, the corresponding graph will be saved with the given name of the file.
            save1 is for CV curve ; and save2 for refit beta-solution.
        formulation (Formulation) : object containing the info about the formulation of the minimization problem we solve.
        time (float) : running time of this action.

    """

    def __init__(self, matrices, param, formulation, numerical_method, label):
        t0 = time()

        # Formulation choosing
        if param.formulation == "not specified":
            param.formulation = formulation
        if param.numerical_method == "not specified":
            param.numerical_method = numerical_method
        name_formulation = param.formulation.name()

        rho = param.formulation.rho
        rho_classification = param.formulation.rho_classification
        e = param.formulation.e
        # Algorithmic method choosing
        numerical_method = choose_numerical_method(
            param.numerical_method, "CV", param.formulation
        )
        param.numerical_method = numerical_method

        if param.lambdas is None:
            if param.logscale:
                param.lambdas = np.array(
                    [param.lamin ** (i / (param.Nlam - 1)) for i in range(param.Nlam)]
                )
            else:
                np.linspace(1.0, param.lamin, param.Nlam)

        # Compute the solution and is the formulation is concomitant, it also compute sigma
        (out, self.yGraph, self.standard_error, self.index_min, self.index_1SE,) = CV(
            matrices,
            param.Nsubset,
            typ=name_formulation,
            num_meth=numerical_method,
            lambdas=param.lambdas,
            seed=param.seed,
            rho=rho,
            rho_classification=rho_classification,
            oneSE=param.oneSE,
            e=e,
            w=param.formulation.w,
            intercept=param.formulation.intercept,
        )

        self.xGraph = param.lambdas
        self.lambda_1SE = param.lambdas[self.index_1SE]
        self.lambda_min = param.lambdas[self.index_min]
        self.formulation = param.formulation

        if self.formulation.concomitant:
            self.beta, self.sigma = out
        else:
            self.beta = out

        self.selected_param = (
            abs(self.beta) > 1e-5
        )  # boolean array, false iff beta_i =0
        self.refit = min_LS(
            matrices, self.selected_param, intercept=self.formulation.intercept
        )
        self.time = time() - t0
        self.save1 = False
        self.save2 = False
        self.label = label

    def __repr__(self):

        string = "\n CROSS VALIDATION : "
        d = len(self.refit)
        selected = self.selected_param[:]
        # this trick is done to plot only selected parameters, excluding intercept
        if self.formulation.intercept:
            selected[0] = False
            string += "\n Intercept : " + str(self.refit[0])

        self.graphic(save=self.save1,logscale=param.logscale)

        nb_select = sum(selected)
        if nb_select > 10:
            top = np.argpartition(abs(self.refit[selected]), -10)[-10:]
            top = np.sort(top)
        else:
            top = np.arange(nb_select)

        plt.bar(range(nb_select), self.refit[selected])
        plt.title(CV_beta["title"])
        plt.xlabel(CV_beta["xlabel"]), plt.ylabel(CV_beta["ylabel"])
        plt.xticks(top, self.label[selected][top], rotation=90)

        if type(self.save2) == str:
            plt.savefig(self.save2)
        plt.show()

        string += "\n   Selected variables :  "
        for i in np.where(self.selected_param)[0]:
            string += self.label[i] + "    "

        string += "\n   Running time :  " + str(round(self.time, 3)) + "s"
        return string

    def graphic(self, se_max=None, save=None, logscale=True, errorevery=5):
        """Method to plot the graphic showing mean squared error over along lambda path once cross validation is computed.

        Args:
            ratio_mse_max (float): float thanks to which the graphic will not show the lambdas from which MSE(lambda)> min(MSE) + ratio * Standard_error(lambda_min).
                this parameter is useful to plot a graph that zooms in the interesting part.
                Default value : None
            logScale (bool) : input that tells to plot the mean square error as a function of lambda, or log10(lambda)
                Default value : True
            errorevery (int) : parameter input of matplotlib.pyplot.errorbar that gives the frequency of the error bars appearence.
                Default value : 5
            save (string) : path to the file where the figure should be saved. If None, then the figure will not be saved.
                Default value : None

        """

        i_min, i_1SE = self.index_min, self.index_1SE

        jmin, jmax = 0, len(self.yGraph) - 1
        if se_max is not None:
            # jmin and jmax are the bounds on the xaxis to know where to zoom in in the lambda path
            y_max = self.yGraph[i_min] + se_max * self.standard_error[i_min]
            while jmin < i_1SE and self.yGraph[jmin] > y_max:
                jmin += 1
            while jmax > i_min and self.yGraph[jmax] > y_max:
                jmax -= 1

        if logscale:
            plt.errorbar(
                np.log10(self.xGraph[jmin : jmax + 1]),
                self.yGraph[jmin : jmax + 1],
                self.standard_error[jmin : jmax + 1],
                label="mean over the k groups of data",
                errorevery=errorevery,
            )
            plt.xlabel(r"$ \log_{10} \lambda / \lambda_{max}$")
            plt.axvline(
                x=np.log10(self.xGraph[i_min]),
                color="k",
                label=r"$\lambda$ (min MSE)",
            )
            plt.axvline(
                x=np.log10(self.xGraph[i_1SE]),
                color="r",
                label=r"$\lambda$ (1SE) ",
            )
        else:
            plt.errorbar(
                self.xGraph[jmin : jmax + 1],
                self.yGraph[jmin : jmax + 1],
                self.standard_error[jmin : jmax + 1],
                label="mean over the k groups of data",
                errorevery=errorevery,
            )
            plt.xlabel(r"$\lambda / \lambda_{max}$")
            plt.axvline(x=self.xGraph[i_min], color="k", label=r"$\lambda$ (min MSE)")
            plt.axvline(x=self.xGraph[i_1SE], color="r", label=r"$\lambda$ (1SE) ")
        plt.title(CV_graph["title"])
        if self.formulation.classification:
            plt.ylabel(CV_graph["ylabel_classification"])
        else:
            plt.ylabel(CV_graph["ylabel"])
        plt.legend()
        if save is not None and type(save)==str:
            plt.savefig(save)
        plt.show()


# Here, the main function used is stability ; from the file stability selection
class solution_StabSel:
    """Class that contains  characteristics of the stability selection computed,
    which also contains a representation method that plot the selected parameters,
    the solution of the not-sparse problem on the selected variables set,
    and the stability plot.

    Attributes:
        distribution (array) : d array of stability ratios.
        lambdas_path (array or string) : for 'first' method : Nlam array of the lambdas used. Other cases : 'not used'.
        distribution_path (array or string) : for 'first' method :  Nlam x d array with stability ratios as a function of lambda.
        threshold (float) : threshold for StabSel, ie for a variable i, stability ratio that is needed to get selected.
        save1,save2 (bool or string) : if a string is given, the corresponding graph will be saved with the given name of the file.
            save1 is for stability plot ; and save2 for refit beta-solution.
        selected_param (numpy.ndarray) : boolean arrays of size d with True when the variable is selected.
        to_label (numpy.ndarray) : boolean arrays of size d with True when the name of the variable should be seen on the graph.
        refit (numpy.ndarray) : solution beta after solving unsparse problem over the set of selected variables.
        formulation (Formulation) : object containing the info about the formulation of the minimization problem we solve.
        time (float) : running time of this action.

    """

    def __init__(self, matrices, param, formulation, numerical_method, label):
        t0 = time()

        # Formulation choosing
        if param.formulation == "not specified":
            param.formulation = formulation
        if param.numerical_method == "not specified":
            param.numerical_method = numerical_method
        name_formulation = param.formulation.name()

        rho = param.formulation.rho
        rho_classification = param.formulation.rho_classification
        e = param.formulation.e
        # Compute the theoretical lam if necessary
        if param.lam == "theoretical":
            lam = param.theoretical_lam
        else:
            lam = param.lam

        # Algorithmic method choosing
        numerical_method = choose_numerical_method(
            param.numerical_method,
            "StabSel",
            param.formulation,
            StabSelmethod=param.method,
            lam=lam,
        )
        param.numerical_method = numerical_method

        # verify the method
        if param.method not in ["first", "max", "lam"]:
            raise ValueError(
                "name of the stability selection method should be one of those : 'first' , 'max' , 'lam'     not {}".format(
                    param.method
                )
            )

        # Compute the distribution
        output = stability(
            matrices,
            StabSelmethod=param.method,
            numerical_method=numerical_method,
            lamin=param.lamin,
            lam=lam,
            Nlam=param.Nlam,
            q=param.q,
            B=param.B,
            percent_nS=param.percent_nS,
            formulation=name_formulation,
            seed=param.seed,
            rho=rho,
            rho_classification=rho_classification,
            true_lam=not param.rescaled_lam,
            e=e,
            w=param.formulation.w,
            intercept=param.formulation.intercept,
        )

        if param.method == "first":
            distribution, distribution_path, lambdas = output
        else:
            distribution, distribution_path, lambdas = (
                output,
                "not computed",
                "not used",
            )

        self.distribution = distribution
        self.distribution_path = distribution_path
        self.lambdas_path = lambdas
        self.selected_param = self.distribution > param.threshold
        self.threshold_label = param.threshold_label
        self.threshold = param.threshold
        self.refit = min_LS(
            matrices,
            self.selected_param,
            intercept=param.formulation.intercept,
        )
        self.save1 = False
        self.save2 = False
        self.method = param.method
        self.formulation = name_formulation
        self.label = label
        self.time = time() - t0

    def __repr__(self):

        string = "\n STABILITY SELECTION : "

        d = len(self.distribution)
        if d > 20:
            top = np.argpartition(self.distribution + np.random.randn(d) * 1e-5, -20)[
                -20:
            ]
            top = np.sort(top)
        else:
            top = np.arange(d)

        self.to_label = self.distribution > self.threshold_label
        D, Dpath, selected = (
            self.distribution[top],
            self.distribution_path,
            np.array(self.selected_param)[top],
        )
        unselected = [not i for i in selected]
        Dselected, Dunselected = np.zeros(len(D)), np.zeros(len(D))
        Dselected[selected], Dunselected[unselected] = (
            D[selected],
            D[unselected],
        )

        plt.bar(
            range(len(Dselected)),
            Dselected,
            color="r",
            label="selected coefficients",
        )
        plt.bar(
            range(len(Dunselected)),
            Dunselected,
            color="b",
            label="unselected coefficients",
        )
        plt.axhline(
            y=self.threshold,
            color="g",
            label="Threshold : thresh = " + str(self.threshold),
        )

        plt.xticks(
            ticks=np.where(self.to_label[top])[0],
            labels=self.label[top][self.to_label[top]],
            rotation=30,
        )
        plt.xlabel(StabSel_graph["xlabel"]), plt.ylabel(
            StabSel_graph["ylabel"]
        ), plt.title(
            StabSel_graph["title"] + self.method + " using " + self.formulation
        ), plt.legend()

        if type(self.save1) == str:
            plt.savefig(self.save1)

        plt.show()
        plot_path = False  # do not plot this graphic for now because it is confusing
        if type(Dpath) != str and plot_path:
            lambdas = self.lambdas_path
            N = len(lambdas)
            for i1, i2 in enumerate(top):
                if selected[i1]:
                    c = "r"
                else:
                    c = "b"
                plt.plot(lambdas, [Dpath[j][i2] for j in range(N)], c)
            p1 = mpatches.Patch(color="red", label="selected coefficients")
            p2 = mpatches.Patch(color="blue", label="unselected coefficients")
            p3 = mpatches.Patch(
                color="green",
                label="Threshold : thresh = " + str(self.threshold),
            )
            plt.legend(handles=[p1, p2, p3], loc=1)
            plt.axhline(y=self.threshold, color="g")
            plt.xlabel(StabSel_path["xlabel"])
            plt.ylabel(StabSel_path["ylabel"])
            plt.title(
                StabSel_path["title"] + self.method + " using " + self.formulation
            )
            plt.show()

        plt.bar(range(len(self.refit[top])), self.refit[top])
        plt.xlabel(StabSel_beta["xlabel"]), plt.ylabel(
            StabSel_beta["ylabel"]
        ), plt.title(StabSel_beta["title"])
        plt.xticks(
            np.where(self.selected_param[top])[0],
            self.label[top][self.selected_param[top]],
            rotation=30,
        )
        if type(self.save2) == str:
            plt.savefig(self.save2)
        plt.show()

        string += "\n   Selected variables :  "
        for i in np.where(selected)[0]:
            string += self.label[i] + "    "

        string += "\n   Running time :  " + str(round(self.time, 3)) + "s"
        return string


# Here, the main function used is Classo ; from the file compact_func
class solution_LAMfixed:
    """Class that contains  characteristics of the lasso computed
    which also contains a representation method that plot this solution.

    Attributes:
        lambdamax (float) : lambda maximum for which the solution is non-null.
        rescaled_lam (bool) : if True, the problem had been computed for lambda*lambdamax (so lambda should be between 0 and 1).
        lambda (float) : lambda for which the problem is solved.
        beta (numpy.ndarray) : solution beta of classo.
        sigma (float) : solution sigma of classo when formulation is 'R2' or 'R4'.
        selected_param (numpy.ndarray) : boolean arrays of size d with True when the variable is selected (which is the case when the i-th component solution of the classo is non-null).
        refit (numpy.ndarray) : solution beta after solving unsparse problem over the set of selected variables.
        formulation (Formulation) : object containing the info about the formulation of the minimization problem we solve.
        time (float) : running time of this action.

    """

    def __init__(self, matrices, param, formulation, numerical_method, label):
        t0 = time()
        self.formulation = formulation
        # Formulation choosing
        if param.formulation == "not specified":
            param.formulation = formulation
        if param.numerical_method == "not specified":
            param.numerical_method = numerical_method
        name_formulation = param.formulation.name()

        rho = param.formulation.rho
        rho_classification = param.formulation.rho_classification
        e = param.formulation.e
        # Compute the theoretical lam if necessary
        if param.lam == "theoretical" or param.lam < 0:
            self.lam = param.theoretical_lam
        else:
            self.lam = param.lam

        # Algorithmic method choosing
        numerical_method = choose_numerical_method(
            param.numerical_method, "LAM", param.formulation, lam=self.lam
        )
        param.numerical_method = numerical_method
        self.rescaled_lam = param.rescaled_lam

        # Compute the solution and is the formulation is concomitant, it also compute sigma
        out = Classo(
            matrices,
            self.lam,
            typ=name_formulation,
            meth=numerical_method,
            rho=rho,
            get_lambdamax=True,
            true_lam=not self.rescaled_lam,
            e=e,
            rho_classification=rho_classification,
            w=param.formulation.w,
            intercept=param.formulation.intercept,
        )

        if param.formulation.concomitant:
            self.lambdamax, self.beta, self.sigma = out
        else:
            self.lambdamax, self.beta = out
        if param.threshold is None:
            param.threshold = np.mean(abs(self.beta))

        self.selected_param = abs(self.beta) > param.threshold
        self.refit = min_LS(
            matrices,
            self.selected_param,
            intercept=param.formulation.intercept,
        )
        self.label = label
        self.time = time() - t0
        self.save = False

    def __repr__(self):

        string = "\n LAMBDA FIXED : "
        d = len(self.beta)
        if d > 20:
            top = np.argpartition(abs(self.beta) + np.random.randn(d) * 1e-5, -20)[-20:]
            top = np.sort(top)
        else:
            top = np.arange(d)

        plt.bar(range(len(top)), self.beta[top]), plt.title(
            LAM_beta["title"] + str(round(self.lam, 3))
        ), plt.xlabel(LAM_beta["xlabel"]), plt.ylabel(LAM_beta["ylabel"])
        plt.xticks(
            np.where(self.selected_param[top])[0],
            self.label[top][self.selected_param[top]],
            rotation=30,
        )

        if type(self.save) == str:
            plt.savefig(self.save)
        plt.show()

        if self.formulation.concomitant:
            string += "\n   Sigma  =  " + str(round(self.sigma, 3))

        string += "\n   Selected variables :  "
        for i in np.where(self.selected_param)[0]:
            string += self.label[i] + "    "

        string += "\n   Running time :  " + str(round(self.time, 3)) + "s"
        return string


def choose_numerical_method(method, model, formulation, StabSelmethod=None, lam=None):
    """Annex function in order to choose the right numerical method, if the given one is invalid.
    In general, it will choose one of the possible optimization scheme for a given formulation.
    When several computation modes are possible, the rules are as follow :

    If possible, always use "Path-Alg", except for fixed lambdas smaller than 0.05
    and for R4 where Path-Alg does not compute the path (paradoxically).

    Else, it uses "DR".

    Args:
        method (str) : input method that is possibly wrong and should be changed.
        If the method is valid for this formulation, it will not be changed.
        model (str) : Computation mode. Can be "PATH", "StabSel", "CV" or "LAM".
        formulation (Formulation) : object containing the info about the formulation of the minimization problem we solve.
        StabSelmethod (str, optional) : if model is "StabSel", it can be "first" , "lam" or "max".
        lam (float, optional) : value of lam (fractional L1 penalty).

    Returns :
        str : method that should be used. Can be "Path-Alg", "DR", "P-PDS" or "PF-PDS"

    """

    if formulation.classification:
        return "Path-Alg"

    # cases where we use classo at a fixed lambda
    elif (model == "LAM") or (model == "StabSel" and StabSelmethod == "lam"):

        if formulation.concomitant:
            if method not in ["Path-Alg", "DR"]:
                if lam > 0.05:
                    return "Path-Alg"
                else:
                    return "DR"

        else:
            if method not in ["Path-Alg", "DR", "P-PDS", "PF-PDS"]:
                if lam > 0.05:
                    return "Path-Alg"
                else:
                    return "DR"

    # cases where we use pathlasso
    else:
        if formulation.classification:
            if method not in ["Path-Alg", "DR", "P-PDS"]:
                return "Path-Alg"

        elif formulation.concomitant:
            if method not in ["Path-Alg", "DR"]:
                if formulation.huber:
                    return "DR"
                else:
                    return "Path-Alg"

        else:
            if method not in ["Path-Alg", "DR", "P-PDS", "PF-PDS"]:
                return "Path-Alg"

    return method


CV_beta = {
    "title": r"Refitted coefficients after CV model selection",
    "xlabel": r"Coefficient index $i$",
    "ylabel": r"Coefficients $\beta_i$ ",
}
CV_graph = {
    "title": r" ",
    "xlabel": r"$\lambda / \lambda_{max}$",
    "ylabel": r"Mean-Squared Error (MSE) ",
    "ylabel_classify": r"Miss classification rate "
}
LAM_beta = {
    "title": r"Coefficients at $\lambda$ = ",
    "xlabel": r"Coefficient index $i$",
    "ylabel": r"Coefficients $\beta_i$ ",
}
PATH_beta_path = {
    "title": r"Coefficients across $\lambda$-path using ",
    "xlabel": r"$\lambda$",
    "ylabel": r"Coefficients $\beta_i$ ",
}
PATH_sigma_path = {
    "title": r"Scale estimate across $\lambda$-path using ",
    "xlabel": r"$\lambda$",
    "ylabel": r"Scale $\sigma$ ",
}
StabSel_graph = {
    "title": r"Stability selection profile of type ",
    "xlabel": r"Coefficient index $i$",
    "ylabel": r"Selection probability ",
}
StabSel_path = {
    "title": r"Stability selection profile across $\lambda$-path using ",
    "xlabel": r"$\lambda$",
    "ylabel": r"Selection probability ",
}
StabSel_beta = {
    "title": r"Refitted coefficients after stability selection",
    "xlabel": r"Coefficient index $i$",
    "ylabel": r"Coefficients $\beta_i$ ",
}



import numpy as np
import numpy.linalg as LA
from .solve_R3 import problem_R3, Classo_R3

"""
Problem    :   min h_rho((Ab - y)/sigma)sigma + simga + lambda ||b||1 with C.b= 0, sigma>0

Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda.
The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type',
which is defined bellow in order to contain all the important parameters of the problem.
One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo).
We solve the problem without normalizing anything.
"""


def Classo_R4(pb, lam):
    pb_type = pb.type  # can be 'Path-Alg' or 'DR'
    (m, d, k), (A, C, y) = pb.dim, pb.matrix
    lamb = lam * pb.lambdamax
    rho = pb.rho
    regpath = pb.regpath

    if lam == 0.0:
        pb_type = "Path-Alg"  # in this case, we will simply use R3.

    # Only alternative to 2prox : one can use the other formulation of the problem which shows that
    #  we can augment the data and then simply solve a concomitant problem
    # (we do that with the method ODE for example becasue it is pretty efficient).
    # Unfortunately we can do that only for fixed lambda and not for any path algorithms
    # because the augmentation of the data required depends on lambda.
    if pb_type == "Path-Alg":
        # trick of mean-shift formulation explained in the pdf "concomitant huber"
        # problem of e ==> same trick to do as explained as in the end of the file compact_func, with r = np.sqrt(2)

        A_aug = np.sqrt(2) * np.concatenate((A, lamb / (2 * rho) * np.eye(m)), axis=1)
        C_aug = np.concatenate((C, np.zeros((k, m))), axis=1)
        y_aug = np.sqrt(2) * y

        if pb.intercept:
            A_aug = A_aug[:, 1:]
            C_aug = C_aug[:, 1:]
            Abar = np.mean(A_aug, axis=0)
            ybar = np.mean(y_aug)
            A_aug = A_aug - Abar
            y_aug = y_aug - ybar

        pb_aug = problem_R3((A_aug, C_aug, y_aug), "Path-Alg")
        beta_aug, s = Classo_R3(pb_aug, lamb / pb_aug.lambdamax)
        s = s / 2
        beta = beta_aug[:-m]
        if pb.intercept:
            betaO = ybar - np.vdot(Abar, beta)
            beta = np.array([betaO] + list(beta))
        return beta, s

    # Else, we do simply doulgas rachford.
    # Hence, the prox is not so easy to compute because there is a root of polynomial of degree 3 to compute.
    # We do that in the function prox_phi_2 which use the function prox_phi_i (prox of one componant),
    # and it uses calc_Newton which uses newton's method with good initialization.

    if not regpath:
        pb.compute_param()

    proj_sigm = pb.proj_sigm
    QA = pb.QA
    Q1 = pb.Q1
    Q2 = pb.Q2
    Proj = pb.Proj
    Anorm = pb.Anorm

    tol = pb.tol * LA.norm(y)  # tolerance rescaled
    gamma = LA.norm(y) * pb.gam / (Anorm ** 2)
    # two vectors usefull to compute the prox of f(b)= sum(wi |bi|)
    w = lamb * gamma * pb.weights
    zerod = np.zeros(d)
    mu, c = pb.mu, pb.c
    root = [0.0] * len(y)
    xs, nu, o, xbar, x = pb.init

    # 2prox
    if pb_type == "DR":
        for i in range(pb.N):
            nv_b = x + Q1.dot(o) - QA.dot(x) - Q2.dot(x - xbar)
            nv_s = (xs + nu) / 2
            if i > 0 and LA.norm(b - nv_b) * Anorm + LA.norm(s - nv_s) < 2 * tol:
                if regpath:
                    return (
                        b,
                        (xs, nu, o, xbar, x),
                        sum(s) / len(s) / pb.sigmax,
                    )
                else:
                    return (b, sum(s) / len(s))

            s, b = nv_s, nv_b
            Ab = A.dot(b)
            p1, p2, root = prox_phi_2(xs, 2 * Ab - o - y, gamma / c, root, rho)

            sup1 = proj_sigm(nu) - s
            sup2 = p1 - s
            sup3 = p2 + y - Ab
            sup4 = prox(2 * b - xbar, w, zerod) - b
            sup5 = Proj.dot(2 * b - x) - b

            xs = xs + mu * sup1
            nu = nu + mu * sup2
            o = o + mu * sup3
            xbar = xbar + mu * sup4
            x = x + mu * sup5

            if LA.norm(b) + LA.norm(s) > 1e6:
                raise ValueError("The algorithm of Doulgas Rachford diverges")

        raise ValueError(
            "The algorithm of Doulgas Rachford did not converge after %i iterations "
            % pb.N
        )

    print("none of the cases ! ")


"""
This function compute the the solution for a given path of lam :
by calling the function 'algo' for each lambda with warm start,
or wuth the method ODE, by computing the whole path thanks to the ODE
that rules Beta and the subgradient s, and then to evaluate it in the given finite path.
"""


def pathlasso_R4(pb, path, n_active=False):
    n = pb.dim[0]
    BETA, SIGMA, tol = [], [], pb.tol
    pb.type = "DR"
    save_init = pb.init
    pb.regpath = True
    pb.compute_param()
    if type(n_active) == int and n_active > 0:
        n_act = n_active
    else:
        n_act = n

    for lam in path:
        X = Classo_R4(pb, lam)
        BETA.append(X[0]), SIGMA.append(X[2])
        pb.init = X[1]

        if (
            sum([(abs(X[0][i]) > 1e-1) for i in range(len(X[0]))]) >= n_act
            or type(X[1]) == str
        ):
            pb.init = save_init
            BETA.extend([BETA[-1]] * (len(path) - len(BETA)))
            SIGMA.extend([SIGMA[-1]] * (len(path) - len(SIGMA)))
            pb.regpath = False
            return (BETA, SIGMA)

    pb.init = save_init
    pb.regpath = False
    return (BETA, SIGMA)


"""
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
"""


class problem_R4:
    def __init__(self, data, algo, rho, intercept=False):
        self.N = 500000

        (AA, C, y) = data
        A = AA[:, :]
        self.weights = np.ones(A.shape[1])
        self.intercept = intercept
        if intercept:
            # add a column of 1 in A, and change weight.
            A = np.concatenate([np.ones((len(A), 1)), A], axis=1)
            C = np.concatenate([np.ones((len(C), 1)), C], axis=1)
            self.weights = np.concatenate([[0.0], self.weights])
            # not exactly what it should be...
            yy = y - np.mean(y)

        self.dim = (
            A.shape[0],
            A.shape[1],
            C.shape[0],
        )
        self.matrix = (A, C, y)

        (m, d, k) = self.dim
        self.tol = 1e-3

        self.regpath = False
        self.name = algo + " Concomitant Huber"
        self.type = algo  # type of algorithm used
        self.rho = rho
        self.mu = 1.95

        self.c = (
            d / LA.norm(A, 2)
        ) ** 2  # parameter for Concomitant problem : the matrix is scaled as c*A^2
        self.gam = np.sqrt(d)

        if not intercept:
            yy = y
        # sigmax = LA.norm(yy)/np.sqrt(m)
        self.sigmax = find_sigmax(yy, rho, m)
        self.lambdamax = 2 * LA.norm(
            (AA.T).dot(h_prime(yy / self.sigmax, rho)), np.infty
        )

        self.init = (
            self.sigmax * np.ones(m),
            self.sigmax * np.ones(m),
            np.zeros(m),
            np.zeros(d),
            np.zeros(d),
        )

    def compute_param(self):
        (A, C, y) = self.matrix
        m, d, k = self.dim
        self.Anorm = LA.norm(A, "fro")
        self.Proj = proj_c(C, d)  # Proj = I - C^t . (C . C^t )^-1 . C
        self.Q1, self.Q2 = QQ(self.c, A)
        self.QA = self.Q1.dot(A)
        self.proj_sigm = lambda vect: (
            [max(0, sum(vect)) / len(vect)] * len(vect)
        )  # here,
        # compared to the Muller&Combettes paper, there is a projection more on sigma =0,
        # for numerical issues...


"""
Functions used in the algorithms, modules needed:
import numpy as np
import numpy.linalg as LA
from .../class_of_problem import problem
"""


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b, w, zeros):
    return np.minimum(b + w, zeros) + np.maximum(b - w, zeros)


# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M, d):
    if LA.matrix_rank(M) == 0:
        return np.eye(d)
    return np.eye(d) - LA.multi_dot([M.T, np.linalg.inv(M.dot(M.T)), M])


def QQ(coef, A):
    return (
        coef * (A.T).dot(LA.inv(2 * np.eye(A.shape[0]) + coef * A.dot(A.T))),
        LA.inv(2 * np.eye(A.shape[1]) + coef * (A.T).dot(A)),
    )


# Compute the real positive root of a polynomial of degree 3 in the form : X^3 + a*X - b with Newton method and a warm start (for Comcomitant problem)
def calc_Newton(a, b, root):
    er = root ** 3 + a * root - b
    i = 0
    bound = min(np.cbrt(b), b / a)
    while abs(er) > 1e-6 and i < 20:
        if root < 0.0:
            root, er = 0.0, -b
        elif root > bound:
            root, er = bound, bound ** 3 + a * bound - b
        root = root - er / (3 * root ** 2 + a)
        er = root ** 3 + a * root - b
        i += 1
    if i == 20:
        r = np.roots([1.0, 0.0, a, -b])
        root = np.amax((r[np.isreal(r)]).real)
    return root


def prox_phi_2(sig, u, gamma, warm_start, rho):
    p, q, ws = np.zeros(len(u)), np.zeros(len(u)), np.zeros(len(u))
    for i in range(len(u)):
        p[i], q[i], ws[i] = prox_phi_i(sig[i], u[i], 2 * gamma, warm_start[i], rho)
    return (p, q, ws)


# Each componant of the prox. Explicit formula were given in the Combettes&Muller paper.
# It is the prox of the function (u,s) --> (0.5*h_rho(u/sigma) + 0.5)*sigma
def prox_phi_i(s, u, gamma, root, rho):
    if u == 0.0:
        return (0, 0, root)
    frac = gamma * rho / abs(u)
    term = s + gamma * (rho ** 2 - 1) / 2
    bool1 = frac >= 1
    bool2 = abs(u) ** 2 <= gamma * (gamma - 2 * s)
    bool3 = term <= 0
    bool4 = abs(u) >= rho * s + gamma * rho * (1 + rho ** 2) / 2

    if bool1 and bool2:
        return (0.0, 0.0, root)
    elif bool3 and not bool1:
        return (0.0, u * (1 - frac), root)
    elif not bool3 and bool4:
        return (term, u * (1 - frac), root)
    root = calc_Newton(2 * s / gamma + 1, 2 * abs(u) / gamma, root)
    return (
        s + gamma * (root ** 2 - 1) / 2,
        u - gamma * root * np.sign(u),
        root,
    )


# Compute the derivative of the huber function, particulary useful for the computing of lambdamax
def h_prime(y, rho):
    m = len(y)
    lrho = rho * np.ones(m)
    return np.maximum(lrho, -y) + np.minimum(y - lrho, 0)


# useful for computing lambdamax.
def find_sigmax(y, rho, e):
    m, evol = len(y), True
    F = [True] * m
    if rho > 1:
        while evol:
            evol = False
            s = LA.norm(y[F]) / np.sqrt(e - (m - sum(F)) * rho ** 2)
            cste = rho * s
            for j in range(m):
                if F[j] and y[j] > cste:
                    F[j], evol = False, True
                elif not F[j] and not y[j] > cste:
                    F[j], evol = True, True
        return s
    else:
        print("rho too little ==> sigma is always 0")



import numpy as np
import numpy.linalg as LA

from .solve_R1 import problem_R1, Classo_R1, pathlasso_R1
from .solve_R2 import problem_R2, Classo_R2, pathlasso_R2
from .solve_R3 import problem_R3, Classo_R3, pathlasso_R3
from .solve_R4 import problem_R4, Classo_R4, pathlasso_R4
from .path_alg import solve_path, pathalgo_general, h_lambdamax


"""
Classo and pathlasso are the main functions,
they can call every algorithm acording
to the method and formulation required
"""

# can be 'Path-Alg', 'P-PDS' , 'PF-PDS' or 'DR'


def Classo(
    matrix,
    lam,
    typ="R1",
    meth="DR",
    rho=1.345,
    get_lambdamax=False,
    true_lam=False,
    e=None,
    rho_classification=-0.0,
    w=None,
    intercept=False,
):

    if w is not None:
        matrices = (matrix[0] / w, matrix[1] / w, matrix[2])
    else:
        matrices = matrix

    X, C, y = matrices

    if typ == "R3":
        if intercept:
            # here we use the fact that for R1 and R3,
            # the intercept is simple beta0 = ybar-Xbar .vdot(beta)
            # so by changing the X to X-Xbar and y to y-ybar
            #  we can solve standard problem
            Xbar, ybar = np.mean(X, axis=0), np.mean(y)
            matrices = (X - Xbar, C, y - ybar)

        if meth not in ["Path-Alg", "DR"]:
            meth = "DR"
        if e is None or e == len(matrices[0]) / 2:
            r = 1.0
            pb = problem_R3(matrices, meth)
        else:
            r = np.sqrt(2 * e / len(matrices[0]))
            pb = problem_R3((matrices[0] * r, matrices[1], matrices[2] * r), meth)
        lambdamax = pb.lambdamax
        if true_lam:
            beta, s = Classo_R3(pb, lam / lambdamax)
        else:
            beta, s = Classo_R3(pb, lam)
        s = s / np.sqrt(e)

        if intercept:
            betaO = ybar - np.vdot(Xbar, beta)
            beta = np.array([betaO] + list(beta))

    elif typ == "R4":

        if meth not in ["Path-Alg", "DR"]:
            meth = "DR"
        if e is None or e == len(matrices[0]):
            r = 1.0
            pb = problem_R4(matrices, meth, rho, intercept=intercept)
        else:
            r = np.sqrt(e / len(matrices[0]))
            pb = problem_R4(
                (matrices[0] * r, matrices[1], matrices[2] * r),
                meth,
                rho / r,
                intercept=intercept,
            )

        lambdamax = pb.lambdamax
        if true_lam:
            beta, s = Classo_R4(pb, lam / lambdamax)
        else:
            beta, s = Classo_R4(pb, lam)

    elif typ == "R2":

        if meth not in ["Path-Alg", "P-PDS", "PF-PDS", "DR"]:
            meth = "ODE"
        pb = problem_R2(matrices, meth, rho, intercept=intercept)
        lambdamax = pb.lambdamax
        if true_lam:
            beta = Classo_R2(pb, lam / lambdamax)
        else:
            beta = Classo_R2(pb, lam)

    elif typ == "C2":

        lambdamax = h_lambdamax(
            matrices, rho_classification, typ="C2", intercept=intercept
        )
        if true_lam:
            out = solve_path(matrices, lam / lambdamax, False, rho_classification, "C2")
        else:
            out = solve_path(
                matrices, lam, False, rho_classification, "C2", intercept=intercept
            )
        if intercept:
            beta0, beta = out[0][-1], out[1][-1]
            beta = np.array([beta0] + list(beta))
        else:
            beta = out[0][-1]

    elif typ == "C1":

        lambdamax = h_lambdamax(matrices, 0, typ="C1", intercept=intercept)
        if true_lam:
            out = solve_path(
                matrices, lam / lambdamax, False, 0, "C1", intercept=intercept
            )
        else:
            out = solve_path(matrices, lam, False, 0, "C1", intercept=intercept)
        if intercept:
            beta0, beta = out[0][-1], out[1][-1]
            beta = np.array([beta0] + list(beta))
        else:
            beta = out[0][-1]

    else:  # LS
        if intercept:
            # here we use the fact that for R1 and R3,
            #  the intercept is simple beta0 = ybar-Xbar .vdot(beta)
            #  so by changing the X to X-Xbar and y to y-ybar
            #  we can solve standard problem
            Xbar, ybar = np.mean(X, axis=0), np.mean(y)
            matrices = (X - Xbar, C, y - ybar)

        if meth not in ["Path-Alg", "P-PDS", "PF-PDS", "DR"]:
            meth = "DR"
        pb = problem_R1(matrices, meth)
        lambdamax = pb.lambdamax
        if true_lam:
            beta = Classo_R1(pb, lam / lambdamax)
        else:
            beta = Classo_R1(pb, lam)

        if intercept:
            betaO = ybar - np.vdot(Xbar, beta)
            beta = np.array([betaO] + list(beta))

    if w is not None:
        if intercept:
            beta[1:] = beta[1:] / w
        else:
            beta = beta / w

    if typ in ["R3", "R4"]:
        if get_lambdamax:
            return (lambdamax, beta, s)
        else:
            return (beta, s)
    if get_lambdamax:
        return (lambdamax, beta)
    else:
        return beta


def pathlasso(
    matrix,
    lambdas=False,
    n_active=0,
    lamin=1e-2,
    typ="R1",
    meth="Path-Alg",
    rho=1.345,
    true_lam=False,
    e=None,
    return_sigm=False,
    rho_classification=0.0,
    w=None,
    intercept=False,
):

    Nactive = n_active
    if Nactive == 0:
        Nactive = False
    if type(lambdas) != bool:
        if lambdas[0] < lambdas[-1]:
            lambdass = [
                lambdas[i] for i in range(len(lambdas) - 1, -1, -1)
            ]  # reverse the list if needed
        else:
            lambdass = [lambdas[i] for i in range(len(lambdas))]
    else:
        lambdass = np.linspace(1.0, lamin, 80)

    if w is not None:
        matrices = (matrix[0] / w, matrix[1] / w, matrix[2])
    else:
        matrices = matrix

    X, C, y = matrices

    if typ == "R2":

        pb = problem_R2(matrices, meth, rho, intercept=intercept)
        lambdamax = pb.lambdamax
        if true_lam:
            lambdass = [lamb / lambdamax for lamb in lambdass]
        BETA = pathlasso_R2(pb, lambdass, n_active=Nactive)

    elif typ == "R3":
        if intercept:
            # here we use the fact that for R1 and R3, the intercept is simple beta0 = ybar-Xbar .vdot(beta) so by changing the X to X-Xbar and y to y-ybar we can solve standard problem
            Xbar, ybar = np.mean(X, axis=0), np.mean(y)
            matrices = (X - Xbar, C, y - ybar)
        if e is None or e == len(matrices[0]) / 2:
            r = 1.0
            pb = problem_R3(matrices, meth)
        else:
            r = np.sqrt(2 * e / len(matrices[0]))
            pb = problem_R3((matrices[0] * r, matrices[1], matrices[2] * r), meth)
        lambdamax = pb.lambdamax
        if true_lam:
            lambdass = [lamb / lambdamax for lamb in lambdass]
        BETA, S = pathlasso_R3(pb, lambdass, n_active=Nactive)
        S = np.array(S) / r ** 2
        BETA = np.array(BETA)
        if intercept:
            BETA = np.array([[ybar - Xbar.dot(beta)] + list(beta) for beta in BETA])

    elif typ == "R4":

        if e is None or e == len(matrices[0]):
            r = 1.0
            pb = problem_R4(matrices, meth, rho, intercept=intercept)
        else:
            r = np.sqrt(e / len(matrices[0]))
            pb = problem_R4(
                (matrices[0] * r, matrices[1], matrices[2] * r),
                meth,
                rho / r,
                intercept=intercept,
            )

        lambdamax = pb.lambdamax
        if true_lam:
            lambdass = [lamb / lambdamax for lamb in lambdass]
        BETA, S = pathlasso_R4(pb, lambdass, n_active=Nactive)
        S = np.array(S) / r ** 2
        BETA = np.array(BETA)

    elif typ == "C2":

        lambdamax = h_lambdamax(
            matrices, rho_classification, typ="C2", intercept=intercept
        )
        if true_lam:
            lambdas = [lamb / lambdamax for lamb in lambdass]
        BETA = pathalgo_general(
            matrices,
            lambdass,
            "C2",
            n_active=Nactive,
            rho=rho_classification,
            intercept=intercept,
        )

    elif typ == "C1":

        lambdamax = h_lambdamax(matrices, 0, typ="C1", intercept=intercept)
        if true_lam:
            lambdass = [lamb / lambdamax for lamb in lambdass]
        BETA = pathalgo_general(
            matrices, lambdass, "C1", n_active=Nactive, intercept=intercept
        )

    else:  # R1
        if intercept:
            # here we use the fact that for R1 and R3,
            #  the intercept is simple beta0 = ybar-Xbar .vdot(beta)
            #  so by changing the X to X-Xbar and y to y-ybar
            #  we can solve standard problem
            Xbar, ybar = np.mean(X, axis=0), np.mean(y)
            matrices = (X - Xbar, C, y - ybar)
        pb = problem_R1(matrices, meth)
        lambdamax = pb.lambdamax
        if true_lam:
            lambdass = [lamb / lambdamax for lamb in lambdass]
        BETA = pathlasso_R1(pb, lambdass, n_active=n_active)

        if intercept:
            BETA = np.array([[ybar - Xbar.dot(beta)] + list(beta) for beta in BETA])

    real_path = [lam * lambdamax for lam in lambdass]

    if w is not None:
        if intercept:
            ww = np.array([1] + list(w))
        else:
            ww = w

        BETA = np.array([beta / ww for beta in BETA])

    if typ in ["R3", "R4"] and return_sigm:
        return (np.array(BETA), real_path, S)
    return (np.array(BETA), real_path)



import numpy as np
import numpy.random as rd
from .compact_func import Classo, pathlasso


"""
Here is the function that does stability selection. It returns the distribution as an d-array.

There is three different stability selection methods implemented here : 'first' ; 'max' ; 'lam'

    - 'first' will compute the whole path until q parameters pop.
               It will then look at those paremeters, and repeat it for each subset of sample.
               It this case it will also return distr_path wich is an n_lam x d - array . It is usefull to have it is one want to plot it.

   - 'max' will do the same but it will stop at a certain lamin that is set at 1e-2 * lambdamax here,
              then will look at the q parameters for which the  max_lam (|beta_i(lam)|) is the highest.

  - 'lam' will, for each subset of sample, compute the classo solution at a fixed lambda. That it will look at the q highest value of |beta_i(lam)|.

"""


def stability(
    matrix,
    StabSelmethod="first",
    numerical_method="Path-Alg",
    Nlam=100,
    lamin=1e-2,
    lam=0.1,
    q=10,
    B=50,
    percent_nS=0.5,
    formulation="LS",
    seed=1,
    rho=1.345,
    rho_classification=-1.0,
    true_lam=False,
    e=1.0,
    w=None,
    intercept=False,
):

    rd.seed(seed)
    n, d = len(matrix[2]), len(matrix[0][0])
    if intercept:
        d += 1
    nS = int(percent_nS * n)
    distribution = np.zeros(d)

    lambdas = np.linspace(1.0, lamin, Nlam)

    if StabSelmethod == "first":

        distr_path = np.zeros((Nlam, d))
        for i in range(B):
            subset = build_subset(n, nS)
            submatrix = build_submatrix(matrix, subset)
            # compute the path until n_active = q.
            BETA = np.array(
                pathlasso(
                    submatrix,
                    lambdas=lambdas,
                    n_active=q + 1,
                    lamin=0,
                    typ=formulation,
                    meth=numerical_method,
                    rho=rho,
                    rho_classification=rho_classification,
                    e=e * percent_nS,
                    w=w,
                    intercept=intercept,
                )[0]
            )

            distr_path = distr_path + (abs(BETA) >= 1e-1)
            # to do : output, instead of lambdas, the average aciv
            """
                distr_path(lambda)_i = 1/B number of time where i is  (among the q-first & activated before lambda) 
            """
        distribution = distr_path[-1]
        return (distribution * 1.0 / B, distr_path * 1.0 / B, lambdas)

    elif StabSelmethod == "lam":

        for i in range(B):
            subset = build_subset(n, nS)
            submatrix = build_submatrix(matrix, subset)
            regress = Classo(
                submatrix,
                lam,
                typ=formulation,
                meth=numerical_method,
                rho=rho,
                rho_classification=rho_classification,
                e=e * percent_nS,
                true_lam=true_lam,
                w=w,
                intercept=intercept,
            )
            if type(regress) == tuple:
                beta = regress[0]
            else:
                beta = regress
            qbiggest = biggest_indexes(abs(beta), q)
            for i in qbiggest:
                distribution[i] += 1

    elif StabSelmethod == "max":

        for i in range(B):
            subset = build_subset(n, nS)
            submatrix = build_submatrix(matrix, subset)
            # compute the path until n_active = q, and only take the last Beta
            BETA = pathlasso(
                submatrix,
                n_active=0,
                lambdas=lambdas,
                typ=formulation,
                meth=numerical_method,
                rho=rho,
                rho_classification=rho_classification,
                e=e * percent_nS,
                w=w,
                intercept=intercept,
            )[0]
            betamax = np.amax(abs(np.array(BETA)), axis=0)
            qmax = biggest_indexes(betamax, q)
            for i in qmax:
                distribution[i] += 1

    return distribution * 1.0 / B


"""
Auxilaries functions that are used in the main function which is stability

"""


# returns the list of the q highest componants of an array, using the fact that it is probably sparse.
def biggest_indexes(array, q):
    qbiggest = []
    nonnul = non_nul_indices(array)
    reduc_array = array[nonnul]
    for i1 in range(q):
        if not np.any(nonnul):
            break
        reduc_index = np.argmax(reduc_array)
        index = nonnul[reduc_index]
        if reduc_array[reduc_index] == 0.0:
            break
        reduc_array[reduc_index] = 0.0
        qbiggest.append(index)
    return qbiggest


# return the list of indices where the componant of the array is null
def non_nul_indices(array):
    L = []
    for i in range(len(array)):
        if not (array[i] == 0.0):
            L.append(i)
    return L


# for a certain threshold, it returns the features that should be selected
def selected_param(distribution, threshold, threshold_label):
    selected, to_label = [False] * len(distribution), [False] * len(distribution)
    for i in range(len(distribution)):
        if distribution[i] > threshold:
            selected[i] = True
        if distribution[i] > threshold_label:
            to_label[i] = True
    return (np.array(selected), np.array(to_label))


# submatrices associated to this subset
def build_submatrix(matrix, subset):
    (A, C, y) = matrix
    subA, suby = A[subset], y[subset]
    return (subA, C, suby)


# random subset of [1,n] of size nS
def build_subset(n, nS):
    return rd.permutation(n)[:nS]



from .path_alg import solve_path_Conc
import numpy as np
import numpy.linalg as LA
from .misc_functions import unpenalized

"""
Problem    :   min ||Ab - y||^2/sigma + n/2 sigma + lambda ||b||1 with C.b= 0 and sigma > 0

Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda.
The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type',
which is defined bellow in order to contain all the important parameters of the problem.
One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo).
We solve the problem without normalizing anything.
"""


def Classo_R3(pb, lam):
    pb_type = pb.type  # can be 'Path-Alg' or 'DR'
    (m, d, k), (A, C, y) = pb.dim, pb.matrix
    sigmax = pb.sigmax

    if lam == 0.0:
        beta = unpenalized(pb.matrix)
        sigma = LA.norm(A.dot(beta) - y) / np.sqrt(m / 2)
        return beta, sigma

    # Path algorithm
    # here we compute the path algo until our lambda, and just take the last beta

    # here we use our path algorithm for concomitant problem, and then only takes the last beta.
    # Actually, the function solve_path_Conc has the argument concomitant= 'fix_lam' so it means it will directly stop when it has to.
    # Then we only have to finc the solution between the last beta computed and the one before.
    if pb_type == "Path-Alg":
        (beta1, beta2), (s1, s2), (r1, r2) = solve_path_Conc(
            (A, C, y), lam, lassopath=False
        )
        dr, ds = r1 - r2, s1 - s2
        teta = root_2(
            LA.norm(dr) ** 2 - ds ** 2,
            np.vdot(dr, r2) - s2 * ds,
            LA.norm(r2) ** 2 - s2 ** 2,
        )
        sigma = (s1 * teta + s2 * (1 - teta)) * sigmax
        beta = beta1 * teta + beta2 * (1 - teta)
        return (beta, sigma)

    regpath = pb.regpath
    if not regpath:
        pb.compute_param()

    lamb = lam * pb.lambdamax
    Anorm = pb.Anorm
    tol = pb.tol * LA.norm(y) / Anorm  # tolerance rescaled
    Proj = proj_c(C, d)  # Proj = I - C^t . (C . C^t )^-1 . C
    QA = pb.QA
    Q1 = pb.Q1
    Q2 = pb.Q2
    # Save some matrix products already computed in problem.compute_param()
    gamma = pb.gam / (pb.Anorm2 * lam)  # Normalize gamma
    w = lamb * gamma * pb.weights
    zerod = np.zeros(d)
    # two vectors usefull to compute the prox of f(b)= sum(wi |bi|)
    mu, c, root = pb.mu, pb.c, 0.0
    xs, nu, o, xbar, x = pb.init

    # 2prox
    if pb_type == "DR":
        for i in range(pb.N):
            nv_b = x + Q1.dot(o) - QA.dot(x) - Q2.dot(x - xbar)
            nv_s = (xs + nu) / 2
            if i > 0 and LA.norm(b - nv_b) + LA.norm(s - nv_s) / Anorm < 2 * tol:
                if regpath:
                    return (b, (xs, nu, o, xbar, x), s)
                else:
                    return (b, s)

            s, b = nv_s, nv_b
            Ab = A.dot(b)
            p1, p2, root = prox_phi_1(xs, 2 * Ab - o - y, gamma / c, root)
            sup1 = max(0, nu) - s
            sup2 = p1 - s
            sup3 = p2 + y - Ab
            sup4 = prox(2 * b - xbar, w, zerod) - b
            sup5 = Proj.dot(2 * b - x) - b

            xs = xs + mu * sup1
            nu = nu + mu * sup2
            o = o + mu * sup3
            xbar = xbar + mu * sup4
            x = x + mu * sup5

            if LA.norm(b) + LA.norm(s) > 1e6:
                raise ValueError("The algorithm of Doulgas Rachford diverges")

        raise ValueError(
            "The algorithm of Doulgas Rachford did not converge after %i iterations "
            % pb.N
        )
    print("none of the cases ! ")


"""
This function compute the the solution for a given path of lam :
by calling the function 'algo' for each lambda with warm start,
or wuth the method ODE, by computing the whole path thanks to the ODE that rules Beta and the subgradient s,
and then to evaluate it in the given finite path.
"""


def pathlasso_R3(pb, path, n_active=False):
    n, d, k = pb.dim
    BETA, SIGMA, tol = [], [], pb.tol

    if pb.type == "Path-Alg":
        y = pb.matrix[2]
        sigmax = LA.norm(y)
        X, LAM, R = solve_path_Conc(pb.matrix, path[-1], n_active=n_active)
        LAM.append(path[-1]), X.append(X[-1]), R.append(R[-1])
        beta2, l2, r2, j = X[0], path[0] + 0.1, -y / LA.norm(y), 0
        for lam in path:
            if lam == 0:
                lam = 1e-4
            while (LA.norm(r2) < l2 / lam) and (j < len(LAM)):
                beta1, l1, r1, beta2, l2, r2, j = (
                    beta2,
                    l2,
                    r2,
                    X[j],
                    LAM[j],
                    R[j],
                    j + 1,
                )
            s1, s2 = l1 / lam, l2 / lam
            dr, ds = r1 - r2, s1 - s2
            teta = root_2(
                LA.norm(dr) ** 2 - ds ** 2,
                np.vdot(dr, r2) - s2 * ds,
                LA.norm(r2) ** 2 - s2 ** 2,
            )
            SIGMA.append((s1 * teta + s2 * (1 - teta)) * sigmax)
            BETA.append(beta1 * teta + beta2 * (1 - teta))
        return (BETA, SIGMA)

    save_init = pb.init
    pb.regpath = True
    pb.compute_param()
    if type(n_active) == int and n_active > 0:
        n_act = n_active
    else:
        n_act = n

    for lam in path:
        X = Classo_R3(pb, lam)
        BETA.append(X[0])
        SIGMA.append(X[-1])
        pb.init = X[1]

        if (
            sum([(abs(X[0][i]) > 1e-1) for i in range(len(X[0]))]) >= n_act
            or type(X[1]) == str
        ):
            pb.init = save_init
            BETA.extend([BETA[-1]] * (len(path) - len(BETA)))
            SIGMA.extend([SIGMA[-1]] * (len(path) - len(SIGMA)))
            pb.regpath = False
            return (BETA, SIGMA)

    pb.init = save_init
    pb.regpath = False
    return (BETA, SIGMA)


"""
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
"""


class problem_R3:
    def __init__(self, data, algo):
        self.N = 500000

        (A, C, y) = data
        self.dim = (A.shape[0], A.shape[1], C.shape[0])
        self.matrix = (A, C, y)

        (m, d, k) = self.dim
        self.weights = np.ones(d)
        self.tol = 1e-6

        self.regpath = False
        self.name = algo + " Concomitant"
        self.type = algo  # type of algorithm used

        self.proj_sigm = lambda x: max(x, 0)
        self.mu = 1.95

        self.gam = np.sqrt(d)
        self.Aty = (A.T).dot(y)
        self.sigmax = LA.norm(y) / np.sqrt(m / 2)
        self.lambdamax = 2 * LA.norm(self.Aty, np.infty) / self.sigmax
        self.init = 1.0, 1.0, np.zeros(m), np.zeros(d), np.zeros(d)

    # Here we compute the costful matrices products and inverts in order to compute it only once, which is especially helpful for warmstarts.
    def compute_param(self):
        (A, C, y) = self.matrix
        m, d, k = self.dim
        self.Anorm = LA.norm(A, "fro")
        self.Anorm2 = self.Anorm ** 2
        c = (d / LA.norm(A, 2)) ** 2
        # parameter for Concomitant problem : the matrix is scaled as c*A^2
        self.c = c
        self.Q1, self.Q2 = QQ(c, A)
        self.QA = self.Q1.dot(A)


"""
Functions used in the algorithms, modules needed :
"""


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b, w, zeros):
    return np.minimum(b + w, zeros) + np.maximum(b - w, zeros)


# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M, d):
    if LA.matrix_rank(M) == 0:
        return np.eye(d)
    return np.eye(d) - LA.multi_dot([M.T, np.linalg.inv(M.dot(M.T)), M])


# Compute the real positive root of a polynomial of degree 3 in the form :
#  X^3 + a*X - b with Newton method and a warm start (for Comcomitant problem)
def calc_Newton(a, b, root):
    er = -b
    while abs(er) > 1e-6:
        root = root - er / (3 * root ** 2 + a)
        er = root ** 3 + a * root - b
    return root


def QQ(coef, A):
    # compute QQ = coef A^t (2.I.+coef A A^t )^-1 , (2.I.+coef A^t A)^-1
    return (
        coef * (A.T).dot(LA.inv(2 * np.eye(A.shape[0]) + coef * A.dot(A.T))),
        LA.inv(2 * np.eye(A.shape[1]) + coef * (A.T).dot(A)),
    )


# Return the cost function of some Beta for a given Lasso problem : L_LS = ||y-Ab||2^2  + lambda* ||b||1
def L_LS(pb, lam, sol):
    return LA.norm(
        pb.matrix[0].dot(sol) - pb.matrix[2]
    ) ** 2 + lam * pb.lambdamax * LA.norm(sol, 1)


# Return the prox operator for the function
# phi = ||(y-Ab)||2^2 /sigma + 1/2 * sigma with the warm start on the computation of the root
def prox_phi_1(sig, u, gamma, warm_start):
    l2u = LA.norm(u)
    Po0 = 4 * gamma * sig + l2u ** 2
    if Po0 < 2 * gamma ** 2:
        return (0.0, 0.0, 0.0)
    else:
        root = calc_Newton((4 * sig / gamma + 6.0), 8 * l2u / gamma, warm_start)
        return (
            sig + 0.5 * gamma * (0.5 * root ** 2 - 1.0),
            u * (1 - gamma * root / l2u),
            root,
        )


# Return the positive root in [0,1] of the polynomial aX^2 + 2bX + c if it exists, 1. if not
def root_2(a, b, c):

    if a == 0.0:
        return c / (2 * b)
    root = (-np.sqrt(b ** 2 - a * c) - b) / a
    if root > 1:
        return 1.0
    return root


# The aim :
# solve the equation :
# |dr|^2 teta^2 + 2 (dr.r2) teta  + |r2|^2 =
# |ds|^2 teta^2 + 2 (ds.s2) teta  + |s2|^2

# Which comes from the equation :
# | teta r1 + (1-teta) r2 | = | teta s1 + (1-teta) s2 |

#  Which comes from the equation :
# lam * | teta r1 + (1-teta) r2 | = | teta l1 + (1-teta) l2 |

# lam * | y - X.( teta beta1 + (1-teta)beta2 ) | = | teta l1 + (1-teta) l2 |

# in other word : $lam.|y-X.beta(gamma)| = gamma $

# so find a gamma such that sigma(gamma) * lam = gamma



from .misc_functions import (
    random_data,
    csv_to_np,
    mat_to_np,
    rescale,
    clr,
    theoretical_lam,
    to_zarr,
)  # , tree_to_matrix
from .compact_func import Classo, pathlasso
from .cross_validation import CV
from .stability_selection import stability
from .solver import (
    classo_problem,
    Data,
    Formulation,
    Model_selection,
    PATHparameters,
    CVparameters,
    StabSelparameters,
    LAMfixedparameters,
    Solution,
    solution_PATH,
    solution_CV,
    solution_StabSel,
    solution_LAMfixed,
    choose_numerical_method,
)



"""
The equation we are solving is:

2N.T.dot(N)  beta_dot      C.T v_dot = s     on active variables (beta_dot elsewhere)

        C    beta_dot      0   v_dot = 0


    and
    (lambda s)_dot  = N.T.N beta_dot   + C.T v_dot              on inactive variables




N = A[F,:]
if intercept :
    N = A[F,:]-mean(A[F,:],axis=0)

"""

import numpy as np
import numpy.linalg as LA

N = 10000
N_frac = 100


class parameters_for_update:
    """Object that has parameters needed at each breaking point during path algorithm

    Attributes :
        number_act  (int) : number of active parameter
        idr         : saves the independant rows of the matrix C resctricted to the actives parameters
        Xt          :  inverse of M
        activity    :  list of boolean, activity[i] is True when variable i is active
        beta        : current solution beta
        s           : current subgradient
        lam         : current lam
        lambdamax   : lambdamax
        M           : matrix to invert
        y           : output
        r           : residual
        F           : F is the set where r<1 and if huber, then it is the set where rho<r<1
        rho         : only use when doing huber path algo


    """

    def __init__(self, matrices, lamin, rho, typ, eps_L2=1e-3, intercept=False):
        if typ == "C2" and rho > 1:
            raise ValueError(
                "For huberized hinge, rho has to be smaller than 1, but here it is :",
                rho,
            )

        (self.A, self.C, self.y) = matrices
        self.lamin = lamin
        self.rho = rho
        self.formulation = typ
        self.intercept = intercept
        n, d, k = len(self.A), len(self.A[0]), len(self.C)
        self.number_act = 0
        self.eps_L2 = eps_L2
        self.idr = [False] * k
        self.activity = [False] * d
        self.beta = np.zeros(d)

        if typ in ["C1", "C2"]:
            r_func = lambda b0, y: y * b0
            dr = self.y
        else:
            r_func = lambda b0, y: b0 - y
            dr = 1.0

        if intercept:
            self.beta0 = find_beta0(r_func, dr, self.y, rho, typ)
            self.Abar = np.mean(self.A, axis=0)
        else:
            self.beta0 = 0.0

        self.F = find_F(r_func(self.beta0, self.y), rho, typ)
        P = self.A[self.F]
        if intercept:
            P = P - np.mean(P, axis=0)
        AtA = 2 * P.T.dot(P) + eps_L2 * np.eye(d)

        self.r = r_func(self.beta0, self.y)
        s = -2 * self.A.T.dot(dr * h_prime(rho, typ)(self.r))
        self.lambdamax = LA.norm(s, np.inf)
        self.lamin = lamin
        self.s = s / self.lambdamax
        self.lam = 1.0

        for i in range(d):
            if self.s[i] == 1.0 or self.s[i] == -1.0:
                self.activity[i] = True
                self.number_act += 1
                if k > 0:
                    to_ad = next_idr1(self.idr, self.C[:, self.activity])
                    if type(to_ad) == int:
                        self.idr[to_ad] = True

        if k == 0:
            self.M = 2 * AtA
        else:
            self.M = np.concatenate(
                (
                    np.concatenate((2 * AtA, self.C.T), axis=1),
                    np.concatenate((self.C, np.zeros((k, k))), axis=1),
                ),
                axis=0,
            )

        N = self.M[self.activity + self.idr, :][:, self.activity + self.idr]
        try:
            self.Xt = LA.inv(N)
        except LA.LinAlgError:
            self.Xt = LA.inv(N + self.eps_L2 * np.eye(len(N)))


# iteration of the function up to solve the path at each breaking points.
def solve_path(matrices, lamin, n_active, rho, typ, intercept=False):
    """
    This functions will compute the path for all the breaking points :
    beta is a piecewise linear function of lambda, and only value on the breaking points
    is computed here

    Args :
        matrices : A (aka X),C,y the matrices of the problem
        lamin : fraction of lambda that gives one criteria to stop : continue while lambda > lamin * lambda_max
        n_active : another criteria to stop
        rho : only useful for huber-classification
        typ : can be 'R1', 'R3', 'R2','C2' or 'C1'

    Return :
        BETA : list of beta(lambda) for lambda in LAMBDA
        LAMBDA : list of breaking points
    """

    d = len(matrices[0][0])
    param = parameters_for_update(matrices, lamin, rho, typ, intercept=intercept)
    if intercept:
        BETA0 = [param.beta0]
    BETA, LAM = [param.beta], [param.lam]
    if param.lam < lamin:
        if intercept:
            return BETA0, BETA, LAM
        else:
            return BETA, LAM
    for i in range(d * N_frac):
        up(param)
        BETA.append(param.beta), LAM.append(param.lam)
        if intercept:
            BETA0.append(param.beta0)
        if (n_active > 0 and param.number_act >= n_active) or param.lam == lamin:
            if intercept:
                return BETA0, BETA, LAM
            else:
                return BETA, LAM

    raise ValueError(
        "The path algorithm did not finsh after %i iterations " % N,
        "and with lamin=%f" % lamin,
        " and with n_active =" + str(n_active),
    )


def solve_path_Conc(matrices, stop, n_active=False, lassopath=True, true_lam=False):
    """
    This functions will compute the path for all the breaking points :
    beta is a piecewise linear function of lambda, and only value on the breaking points
    is computed here

    Args :
        matrices : A (aka X),C,y the matrices of the problem
        stop : fraction of lambda_R3 that gives one criteria to stop :
            continue while lambda_R3 > stop * lambda_R3_max
            but this is the lambda of R3, which live in another space..
        n_active : another criteria to stop

    Return :
        BETA : list of beta(lambda) for lambda in LAMBDA
        LAM : list of breaking points
        R : list of residuals rescaled
    """

    (A, C, y) = matrices
    n, d, k = len(A), len(A[0]), len(C)
    # to compute r = (A beta - y)/||y|| more efficientely ; and we set reduclam=lam/stop to 2 so that if stop = 0, the condition reduclam < ||r|| is never furfilled
    A_over_NORMy, y_over_NORMy, reduclam = (
        A / (LA.norm(y)),
        y / (LA.norm(y)),
        2.0,
    )
    if lassopath:
        lamin, R = 0, [-y_over_NORMy]
    else:
        lamin, beta_old, reduclam_old, r_old = (
            0,
            np.zeros(d),
            1.0,
            -y_over_NORMy,
        )

    param = parameters_for_update(matrices, lamin, 0, "R3")
    BETA, LAM = [param.beta], [param.lam]
    for i in range(d * N_frac):

        up(param)
        BETA.append(param.beta), LAM.append(param.lam)
        param.r = A_over_NORMy.dot(param.beta) - y_over_NORMy
        if stop != 0:
            reduclam = param.lam / stop
        if lassopath:
            R.append(param.r)
            if (
                (reduclam <= LA.norm(param.r))
                or (param.number_act >= n - k)
                or (n_active > 0 and param.number_act >= n_active)
            ):
                return (BETA, LAM, R)
        else:
            if reduclam <= LA.norm(param.r):
                return (
                    (beta_old, param.beta),
                    (reduclam_old, reduclam),
                    (r_old, param.r),
                )
            beta_old, reduclam_old, r_old = param.beta, reduclam, param.r

    raise ValueError(
        "The concomitant path algorithm did not finsh after %i iterations " % N,
        "and with lamin=" + str(stop),
        " and with n_active =" + str(n_active),
    )


def pathalgo_general(matrix, path, typ, n_active=False, rho=0, intercept=False):
    """
    This function is only to interpolate the solution path between the breaking points
    """
    BETA, BETA0, i = [], [], 0
    if intercept:
        B0, B, sp_path = solve_path(
            matrix, path[-1], n_active, rho, typ, intercept=intercept
        )
        B0.append(B0[-1])
    else:
        B, sp_path = solve_path(
            matrix, path[-1], n_active, rho, typ, intercept=intercept
        )

    sp_path.append(path[-1]), B.append(B[-1])
    for lam in path:
        while lam < sp_path[i + 1]:
            i += 1
        teta = (sp_path[i] - lam) / (sp_path[i] - sp_path[i + 1])
        BETA.append(B[i] * (1 - teta) + B[i + 1] * teta)
        if intercept:
            BETA0.append(B0[i] * (1 - teta) + B0[i + 1] * teta)

    if intercept:
        BETA = np.array([[BETA0[i]] + list(BETA[i]) for i in range(len(BETA0))])

    return BETA


def pathalgo_huber_cl(matrix, path, rho, n_active=False, intercept=False):
    return pathalgo_general(
        matrix, path, "C2", n_active=n_active, rho=rho, intercept=intercept
    )


def pathalgo_cl(matrix, path, n_active=False, intercept=False):
    return pathalgo_general(matrix, path, "C1", n_active, intercept=intercept)


def up(param):
    """
    Function to call to go from a breaking point to the next one
    """
    formulation = param.formulation
    if formulation in ["R1", "R3"]:
        up_LS(param)
    elif formulation == "R2":
        up_huber(param)
    elif formulation == "C1":
        up_cl(param)
    elif formulation == "C2":
        up_huber_cl(param)
    else:
        raise ValueError(
            "Unknown formulation {}, please enter one of those : R1, R2, R3, C1, C2".format()
        )


def up_LS(param):
    """
    Function to call to go from a breaking point to the next one
    """
    # parameters that does not change
    lambdamax = param.lambdamax
    lamin = param.lamin
    M = param.M
    C = param.C
    eps_L2 = param.eps_L2

    # parameters to be updated
    number_act = param.number_act
    idr = param.idr
    Xt = param.Xt
    activity = param.activity
    beta = param.beta
    s = param.s
    lam = param.lam

    d = len(activity)
    L = [lam] * d
    Mat = M[:d, :d]
    beta_dot, lam_s_dot = derivatives(activity, s, Mat, M[d:, :d], Xt, idr, number_act)
    for i in range(d):
        bi, di, e, s0 = beta[i], beta_dot[i], lam_s_dot[i], s[i]
        if activity[i]:
            if abs(bi * di) > 1e-10 and bi * di > 0:
                L[i] = bi / (di * lambdamax)
        else:
            if abs(e - s0) < 1e-10:
                continue
            if e > s0:
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam
    dlamb = min(min(L), lam - lamin)
    # Update matrix inverse, list of rows in C and activity
    for i in range(d):
        if L[i] < dlamb + 1e-10:
            if activity[i]:
                activity[i], number_act = False, number_act - 1
                if len(M) > d:
                    to_ad = next_idr2(idr, C[:, activity])
                    if type(to_ad) == int:
                        idr[to_ad] = False
            else:
                # x = M[:, activity + idr][i]
                # al = M[i, i] - np.vdot(x, Xt.dot(x))
                # if (abs(al) < 1e-10): break
                activity[i], number_act = True, number_act + 1
                if len(M) > d:
                    to_ad = next_idr1(idr, C[:, activity])
                    if type(to_ad) == int:
                        idr[to_ad] = True

    N = M[activity + idr, :][:, activity + idr]

    try:
        Xt = LA.inv(N)
    except LA.LinAlgError:
        Xt = LA.inv(N + eps_L2 * np.eye(len(N)))

    beta = beta - lambdamax * beta_dot * dlamb
    if not (lam == dlamb):
        s = lam_s_dot + lam / (lam - dlamb) * (s - lam_s_dot)
    lam -= dlamb

    param.number_act = number_act
    param.idr = idr
    param.Xt = Xt
    param.activity = activity
    param.beta = beta
    param.s = s
    param.lam = lam


def up_huber(param):
    """
    Function to call to go from a breaking point to the next one
    """

    # parameters that does not change
    lambdamax = param.lambdamax
    lamin = param.lamin
    A = param.A
    y = param.y
    C = param.C
    rho = param.rho
    eps_L2 = param.eps_L2

    # parameters to be updated
    number_act = param.number_act
    idr = param.idr
    Xt = param.Xt
    activity = param.activity
    F = param.F
    beta = param.beta
    s = param.s
    lam = param.lam
    M = param.M
    r = param.r

    d = len(activity)
    L = [lam] * d
    Mat = M[:d, :d]
    beta_dot, lam_s_dot = derivatives(activity, s, Mat, M[d:, :d], Xt, idr, number_act)
    for i in range(d):
        bi, di, e, s0 = beta[i], beta_dot[i], lam_s_dot[i], s[i]
        if activity[i]:
            if abs(bi * di) > 1e-10 and bi * di > 0:
                L[i] = bi / (di * lambdamax)
        else:
            if abs(e + s0) < 1e-10 or abs(s0) > 1:
                continue
            if e > s0:
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam

    dlamb = min(min(L), lam, lam - lamin)
    huber_up = False
    ADl = -A.dot(beta_dot) * lambdamax
    if param.intercept:
        ADl += np.vdot(param.Abar, beta_dot) * lambdamax

    huber_up = False
    j_switch = None
    for j in range(len(r)):
        #     find if there is a  0< dl < dlamb such that F[j] and |r[j]+ADl[j]*dl|>rho
        # or  find if there is a  0< dl < dlamb such that not F[j] |r[j]+ADl[j]*dl|<rho
        if abs(ADl[j]) < 1e-6 or abs(r[j]) < 1e-6 or abs(abs(r[j]) - rho) < 1e-6:
            continue
        c = -r[j] / ADl[j]
        dl = (1 - rho / abs(r[j])) * c
        if dl < 0:
            if c > 0:
                dl = (1 + rho / abs(r[j])) * c
            else:
                dl = dlamb
        if dl < dlamb:
            huber_up, j_switch, dlamb = True, j, dl
    beta = beta - lambdamax * beta_dot * dlamb
    s = lam_s_dot + lam / (lam - dlamb) * (s - lam_s_dot)
    r = r + ADl * dlamb
    lam = lam - dlamb
    if param.intercept:
        beta0_dot = -np.vdot(np.mean(A[F], axis=0), beta_dot)
        beta0 = param.beta0 - lambdamax * beta0_dot * dlamb

    if huber_up:
        F[j_switch] = not F[j_switch]
        if param.intercept:
            P = A[F] - np.mean(A[F], axis=0)
            M[:d, :][:, :d] = 2 * P.T.dot(P) + eps_L2 * np.eye(d)
        else:
            M[:d, :][:, :d] = 2 * A[F].T.dot(A[F]) + eps_L2 * np.eye(d)
    else:
        # Update matrix inverse, list of rows in C and activity
        for i in range(d):
            if L[i] < dlamb + 1e-10:
                if activity[i]:
                    activity[i], number_act = False, number_act - 1
                    if len(M) > d:
                        to_ad = next_idr2(idr, M[d:, :d][:, activity])
                        if type(to_ad) == int:
                            idr[to_ad] = False
                else:
                    activity[i], number_act = True, number_act + 1
                    if len(M) > d:
                        to_ad = next_idr1(idr, M[d:, :d][:, activity])
                        if type(to_ad) == int:
                            idr[to_ad] = True

    N = M[activity + idr, :][:, activity + idr]

    try:
        Xt = LA.inv(N)
    except LA.LinAlgError:
        Xt = LA.inv(N + eps_L2 * np.eye(len(N)))

    param.number_act = number_act
    param.idr = idr
    param.Xt = Xt
    param.activity = activity
    param.F = F
    param.beta = beta
    param.s = s
    param.lam = lam
    param.M = M
    param.r = r
    if param.intercept:
        param.beta0 = beta0


def up_cl(param):
    """
    Function to call to go from a breaking point to the next one
    """

    lambdamax = param.lambdamax
    lamin = param.lamin
    A = param.A
    y = param.y
    eps_L2 = param.eps_L2

    number_act = param.number_act
    idr = param.idr
    Xt = param.Xt
    activity = param.activity
    F = param.F
    beta = param.beta
    s = param.s
    lam = param.lam
    M = param.M
    r = param.r

    d = len(activity)
    L = [lam] * d
    Mat = M[:d, :d]
    beta_dot, lam_s_dot = derivatives(activity, s, Mat, M[d:, :d], Xt, idr, number_act)
    for i in range(d):
        bi, di, e, s0 = beta[i], beta_dot[i], lam_s_dot[i], s[i]
        if activity[i]:
            if abs(bi * di) > 1e-10 and bi * di > 0:
                L[i] = bi / (di * lambdamax)
        else:
            if abs(e + s0) < 1e-10 or abs(s0) > 1:
                continue
            if e > s0:
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam

    dlamb = min(min(L), lam, lam - lamin)
    max_up = False
    j_switch = None
    yADl = -y * A.dot(beta_dot) * lambdamax
    if param.intercept:
        yADl += y * np.vdot(param.Abar, beta_dot) * lambdamax
    for j in range(len(r)):
        #     find if there is a  0< dl < dlamb such that F[j] and r[j]+yADl[j]*dl > 1
        # or  find if there is a  0< dl < dlamb such that not F[j] r[j]+yADl[j]*dl < 1
        if abs(r[j] - 1) < 1e-4:
            continue
        if yADl[j] != 0.0:
            dl = (1 - r[j]) / yADl[j]
        else:
            dl = -1
        if dl < dlamb and dl > 0:
            max_up, j_switch, dlamb = True, j, dl

    beta = beta - lambdamax * beta_dot * dlamb
    s = lam_s_dot + lam / (lam - dlamb) * (s - lam_s_dot)
    r = r + yADl * dlamb
    lam = lam - dlamb
    if param.intercept:
        AbarF = np.mean(A[F], axis=0)
        beta0_dot = -np.vdot(AbarF, beta_dot)
        beta0 = param.beta0 - lambdamax * beta0_dot * dlamb

    if max_up:
        F[j_switch] = not F[j_switch]
        if param.intercept:
            P = A[F] - np.mean(A[F], axis=0)
            M[:d, :][:, :d] = 2 * P.T.dot(P) + eps_L2 * np.eye(d)
        else:
            M[:d, :][:, :d] = 2 * A[F].T.dot(A[F]) + eps_L2 * np.eye(d)
    else:
        # Update matrix inverse, list of rows in C and activity
        for i in range(d):
            if L[i] < dlamb + 1e-10:
                if activity[i]:
                    activity[i], number_act = False, number_act - 1
                    if len(M) > d:
                        to_ad = next_idr2(idr, M[d:, :d][:, activity])
                        if type(to_ad) == int:
                            idr[to_ad] = False
                else:
                    # x = M[:,activity+idr][i]
                    # al = M[i,i]-np.vdot(x,Xt.dot(x))
                    # if (abs(al)<1e-10): break
                    activity[i], number_act = True, number_act + 1
                    if len(M) > d:
                        to_ad = next_idr1(idr, M[d:, :d][:, activity])
                        if type(to_ad) == int:
                            idr[to_ad] = True

    N = M[activity + idr, :][:, activity + idr]

    try:
        Xt = LA.inv(N)
    except LA.LinAlgError:
        Xt = LA.inv(N + eps_L2 * np.eye(len(N)))

    param.number_act = number_act
    param.idr = idr
    param.Xt = Xt
    param.activity = activity
    param.F = F
    param.beta = beta
    param.s = s
    param.lam = lam
    param.M = M
    param.r = r
    if param.intercept:
        param.beta0 = beta0


def up_huber_cl(param):
    """
    Function to call to go from a breaking point to the next one
    """

    # parameters that does not change
    lambdamax = param.lambdamax
    lamin = param.lamin
    A = param.A
    y = param.y
    rho = param.rho
    eps_L2 = param.eps_L2

    # parameters to be updated
    number_act = param.number_act
    idr = param.idr
    Xt = param.Xt
    activity = param.activity
    F = param.F
    beta = param.beta
    s = param.s
    lam = param.lam
    M = param.M
    r = param.r

    d = len(activity)
    L = [lam] * d
    Mat = M[:d, :d]
    beta_dot, lam_s_dot = derivatives(
        activity,
        s,
        Mat,
        M[d:, :d],
        Xt,
        idr,
        number_act,
    )
    for i in range(d):
        bi, di, e, s0 = beta[i], beta_dot[i], lam_s_dot[i], s[i]
        if activity[i]:
            if abs(bi * di) > 1e-10 and bi * di > 0:
                L[i] = bi / (di * lambdamax)
        else:
            if abs(e - s0) < 1e-10 or abs(s0) > 1:
                continue
            if e > s0:
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam

    dlamb = min(min(L), lam, lam - lamin)
    max_up = False
    j_switch = None
    yADl = -y * A.dot(beta_dot) * lambdamax
    if param.intercept:
        yADl += y * np.vdot(param.Abar, beta_dot) * lambdamax
    for j in range(len(r)):
        #     find if there is a  0< dl < dlamb such that F[j] and r[j]+yADl[j]*dl > 1
        # or  find if there is a  0< dl < dlamb such that not F[j] r[j]+yADl[j]*dl < 1
        #     find if there is a  0< dlhuber < dlamb such that F[j] and r[j]+yADl[j]*dl < rho
        # or  find if there is a  0< dlhuber < dlamb such that not F[j] r[j]+yADl[j]*dl > rho

        if yADl[j] != 0.0:
            dlmax = (1 - r[j]) / yADl[j]
            dlhuber = (rho - r[j]) / yADl[j]
        else:
            dlmax, dlhuber = dlamb + 1, dlamb + 1
        if dlmax <= 0.0:
            dlmax = dlamb + 1
        if dlhuber <= 0.0:
            dlhuber = dlamb + 1

        dl = min(dlhuber, dlmax)

        if dl < dlamb:
            max_up, j_switch, dlamb = True, j, dl

    beta = beta - lambdamax * beta_dot * dlamb
    s = lam_s_dot + lam / (lam - dlamb) * (s - lam_s_dot)
    r = r + yADl * dlamb
    lam = lam - dlamb
    if param.intercept:
        beta0_dot = -np.vdot(np.mean(A[F], axis=0), beta_dot)
        beta0 = param.beta0 - lambdamax * beta0_dot * dlamb

    if max_up:
        F[j_switch] = not F[j_switch]
        if param.intercept:
            P = A[F] - np.mean(A[F], axis=0)
            M[:d, :][:, :d] = 2 * P.T.dot(P) + eps_L2 * np.eye(d)
        else:
            M[:d, :][:, :d] = 2 * A[F].T.dot(A[F]) + eps_L2 * np.eye(d)

    else:
        # Update matrix inverse, list of rows in C and activity
        for i in range(d):
            if L[i] < dlamb + 1e-10:
                if activity[i]:
                    activity[i], number_act = False, number_act - 1
                    if len(M) > d:
                        to_ad = next_idr2(idr, M[d:, :d][:, activity])
                        if type(to_ad) == int:
                            idr[to_ad] = False
                else:
                    # x = M[:,activity+idr][i]
                    # al = M[i,i]-np.vdot(x,Xt.dot(x))
                    # if (abs(al)<1e-10): break
                    activity[i], number_act = True, number_act + 1
                    if len(M) > d:
                        to_ad = next_idr1(idr, M[d:, :d][:, activity])
                        if type(to_ad) == int:
                            idr[to_ad] = True

    N = M[activity + idr, :][:, activity + idr]

    try:
        Xt = LA.inv(N)
    except LA.LinAlgError:
        Xt = LA.inv(N + eps_L2 * np.eye(len(N)))

    param.number_act = number_act
    param.idr = idr
    param.Xt = Xt
    param.activity = activity
    param.F = F
    param.beta = beta
    param.s = s
    param.lam = lam
    param.M = M
    param.r = r
    if param.intercept:
        param.beta0 = beta0


# Compute the derivatives of the solution Beta and the derivative of lambda*subgradient thanks to the ODE
def derivatives(activity, s, Mat, C, Inv, idr, number_act):
    """
    Compute the derivatives of the solution Beta and the derivative of lambda*subgradient
    thanks to the following equation :

    2A[active].T A[active] beta_dot + C[active].T v_dot = s       on active variables
    C[active]   beta_dot                               = 0

     (lambda s)_dot  = 2A[inactive].T A[active] beta_dot         on inactive variables

     Then : Mat = A[inactive].T A[active]  and Inv is the inverse of a matrix of the first equaction.

    """
    beta_dot = np.zeros(len(activity))
    beta_dot[activity] = -Inv[:number_act, :number_act].dot(s[activity])
    lam_s_dot = -Mat.dot(beta_dot)

    if len(C) > 0:
        v_dot = np.zeros(len(C))
        v_dot[idr] = Inv[number_act:, :number_act].dot(s[activity])
        lam_s_dot += C.T.dot(v_dot)

    return (beta_dot, lam_s_dot)


# Upddate a list of constraints which are independant if we restrict the matrix C to the acrive set (C_A has to have independant rows)
# When we ad an active parameter
def next_idr1(liste, mat):
    if sum(liste) == len(mat):
        return False
    if sum(liste) == 0:
        for i in range(len(mat)):
            for j in range(len(mat[0])):
                if not (mat[i, j] == 0):
                    return i
        return False
    Q = LA.qr(mat[liste, :].T)[0]
    for j in range(len(mat)):
        if (not liste[j]) and (
            LA.norm(mat[j] - LA.multi_dot([Q, Q.T, mat[j]])) > 1e-10
        ):
            return j
    return False


# When we remove an active parameter
def next_idr2(liste, mat):
    if sum(liste) == 0:
        return False
    R = LA.qr(mat[liste, :].T)[1]
    for i in range(len(R)):
        if abs(R[i, i]) < 1e-10:  # looking for the i-th True element of liste
            j, somme = 0, liste[0]
            while somme <= i:
                j, somme = j + 1, somme + liste[j + 1]
            return j
    return False


# Update the invers of a matrix whom we add a line, which is useul to compute the derivatives
def next_inv(Xt, B, al, ligne):
    n = len(Xt)
    Yt = np.zeros((n + 1, n + 1))
    alpha = 1 / al
    B = np.array([B])
    b1 = Xt[:ligne, :][:, :ligne] + alpha * B[:, :ligne].T.dot(B[:, :ligne])
    b2 = Xt[ligne:, :][:, :ligne] + alpha * B[:, ligne:].T.dot(B[:, :ligne])
    b4 = Xt[ligne:, :][:, ligne:] + alpha * B[:, ligne:].T.dot(B[:, ligne:])
    col1 = np.concatenate((b1, -alpha * B[:, :ligne], b2), axis=0)
    col2 = np.concatenate((b2.T, -alpha * B[:, ligne:], b4), axis=0)
    col = np.concatenate(
        (-alpha * B[0, :ligne], [alpha], -alpha * B[0, ligne:]), axis=0
    )
    return np.concatenate((col1, np.array([col]).T, col2), axis=1)


def h_lambdamax(matrices, rho, typ="R1", intercept=False):
    param = parameters_for_update(matrices, 0.0, rho, typ, intercept=intercept)
    return param.lambdamax


# Compute the derivative of the huber function, particulary useful for the computing of lambdamax
def h_prime(rho, typ):
    if rho > 0 and typ == "R2":
        # huber regress
        # where grad(h)(r) is :
        # r if -rho<r<rho   ; -rho if r < -rho  ; rho if r > rho
        return lambda r: np.minimum(r, rho) + np.maximum(r, -rho) - r

    elif typ == "C2":
        # huber classification
        # where grad(h)(r) is :
        # (r-1) if 1>r>rho   ;  (rho-1) if r<rho  ;  0  if r >1
        # NF*beta0 = (1-rho)*Nminus + np.sum(y[F])
        return lambda r: np.minimum(np.maximum(r, rho) - 1.0, 0.0)

    elif typ == "C1":
        # classification
        # grad(h)(r) is :
        # (r-1) if 1>r  ; 0  if r >1
        return lambda r: np.minimum(r - 1, 0.0)

    else:
        # R1 or R3
        # grad(h)(r) is:
        # r
        return lambda r: r


def find_F(y, rho, typ):
    """
    Find initial value of F, and beta0 if intercept
    F is the set of indices for which objective is quadratic
    """
    if rho > 0 and typ == "R2":
        # huber regress
        return (y > -rho) & (y < rho)
    elif typ == "C2":
        # huber classify
        return (y > rho) & (y < 1)
    elif typ == "C1":
        return y < 1
    else:
        # no huber
        return np.ones(len(y), dtype=bool)


def find_beta0(r, dbeta0, y, rho, typ):
    gradh = lambda b0: np.sum(dbeta0 * h_prime(rho, typ)(r(b0, y)))
    beta0 = binary_search(gradh, min(y), max(y))
    return beta0


def binary_search(f, a, b, tol=1e-8):
    c = (a + b) / 2
    if f(a) * f(b) > 0:
        print("gradh(min(y)) = ", f(a))
        print("gradh(max(y)) = ", f(b))
        raise ValueError("Error in binary search for initial intercept")
    while abs(f(c)) > tol:
        if f(c) * f(a) < 0:
            b = c
        else:
            a = c
        c = (a + b) / 2
    return c


"""
Function that solve the concomitant problem for every lambda thanks to the previous function:
we firstly compute all the non concomitant Least square problems,
then we use it to find sigma and so the solution, using the equation for sigma:

sigma = || A*B(lambda*sigma) - y ||_2       (where B(lambda) is found thanks to solve_path)

            teta = (sp_path[i]-lam)/(sp_path[i]-sp_path[i+1])
"""



import numpy as np
import numpy.random as rd
import numpy.linalg as LA
from .compact_func import Classo, pathlasso

n_lam = 80


def compute_1SE(mse_max, MSE, i):
    j = i
    while j > 0 and MSE[j] < mse_max:
        j -= 1
    return j


def train_test_CV(n, k, test_pourcent):
    idx, training_size = rd.permutation(n), int(n - n * test_pourcent)
    idx_train, idx_test = idx[:training_size], idx[training_size:]
    SUBLIST, end = [], 0
    for i in range(k):
        begin, end = end, end + training_size // k
        if i < training_size % k:
            end += 1
        SUBLIST.append(idx[begin:end])
    return (SUBLIST, idx_train, idx_test)


def train_test_i(SUBLIST, i):
    training_set, test_set = np.array([], dtype=int), SUBLIST[i]
    for j in range(len(SUBLIST)):
        if j != i:
            training_set = np.concatenate((training_set, SUBLIST[j]))
    return (training_set, test_set)


def training(
    matrices,
    typ,
    num_meth,
    training_set,
    rho,
    rho_classification,
    e,
    lambdas,
    w,
    intercept,
):
    (A, C, y) = matrices
    mat = (A[training_set], C, y[training_set])
    sol = pathlasso(
        mat,
        lambdas=lambdas,
        typ=typ,
        meth=num_meth,
        rho=rho,
        e=e,
        rho_classification=rho_classification,
        w=w,
        intercept=intercept,
    )[0]
    return sol


def test_i(
    matrices,
    typ,
    num_meth,
    SUBLIST,
    i,
    rho,
    rho_classification,
    e,
    lambdas,
    w,
    intercept,
):
    training_set, test_set = train_test_i(SUBLIST, i)
    BETA = training(
        matrices,
        typ,
        num_meth,
        training_set,
        rho,
        rho_classification,
        e,
        lambdas,
        w,
        intercept,
    )
    n_lam = len(lambdas)
    residual = np.zeros(n_lam)
    for j in range(n_lam):
        residual[j] = accuracy_func(
            matrices[0][test_set],
            matrices[2][test_set],
            BETA[j],
            typ,
            rho=rho,
            rho_classification=rho_classification,
            intercept=intercept,
        )

    return residual


def average_test(
    matrices,
    typ,
    num_meth,
    SUBLIST,
    rho,
    rho_classification,
    e,
    lambdas,
    w,
    intercept,
):
    k = len(SUBLIST)
    n_lam = len(lambdas)
    RESIDUAL = np.zeros((k, n_lam))
    for i in range(k):
        RESIDUAL[i, :] = test_i(
            matrices,
            typ,
            num_meth,
            SUBLIST,
            i,
            rho,
            rho_classification,
            e,
            lambdas,
            w,
            intercept,
        )
    MSE = np.mean(RESIDUAL, axis=0)
    SE = np.std(RESIDUAL, axis=0) / np.sqrt(k)
    return (MSE, SE)


def CV(
    matrices,
    k,
    typ="R1",
    num_meth="Path-Alg",
    test=0.0,
    seed=1,
    rho=1.345,
    rho_classification=-1.0,
    e=1.0,
    lambdas=None,
    Nlam=100,
    oneSE=True,
    w=None,
    intercept=False,
):

    if lambdas is None:
        lambdas = np.linspace(1.0, 1e-3, Nlam)

    rd.seed(seed)
    (A, C, y) = matrices
    SUBLIST, idx_train, idx_test = train_test_CV(len(y), k, test)
    MSE, SE = average_test(
        matrices,
        typ,
        num_meth,
        SUBLIST,
        rho,
        rho_classification,
        e,
        lambdas,
        w,
        intercept,
    )
    i = np.argmin(MSE)
    i_1SE = compute_1SE(MSE[i] + SE[i], MSE, i)
    if oneSE:
        lam = lambdas[i_1SE]
    else:
        lam = lambdas[i]
    out = Classo(
        (A[idx_train], C, y[idx_train]),
        lam,
        typ=typ,
        meth=num_meth,
        rho=rho,
        e=e,
        rho_classification=rho_classification,
        w=w,
        intercept=intercept,
    )
    return (out, MSE, SE, i, i_1SE)


# Computation of the residual, for huber, LS, huber_classification and classification
def hub(r, rho):
    h = 0
    for j in range(len(r)):
        if abs(r[j]) < rho:
            h += r[j] ** 2
        elif r[j] > 0:
            h += (2 * r[j] - rho) * rho
        else:
            h += (-2 * r[j] - rho) * rho
    return h


def hinge(A, y, beta):
    return sum(np.maximum(0, 1 - y * A.dot(beta)) ** 2)


def huber_hinge(A, y, beta, rho):
    h = np.maximum(0, 1 - y * A.dot(beta))
    s = 0
    for i in range(len(h)):
        if h[i] < rho:
            s += 2 * h[i] * rho - rho ** 2
        else:
            s += h[i] ** 2
    return s

def misclassification_rate(X,y,beta):
    yhat = sign(X.dot(beta))
    return np.sum(y != yhat) / len(y)


def accuracy_func(
    A, y, beta, typ="R1", rho=1.345, rho_classification=-1.0, intercept=False
):

    if intercept:
        Aprime = np.concatenate([np.ones((len(A), 1)), A], axis=1)
    else:
        Aprime = A[:, :]

    n = len(y)

    if typ == "R2":
        return hub(Aprime.dot(beta) - y, rho) / n
    elif typ in ["C1","C2"]:
        return misclassification_rate(Aprime, y, beta)
    else:
        return LA.norm(Aprime.dot(beta) - y, 2) ** 2 / n
